"""Manage `UploadJob`s using json files

An upload "task" is like an item on a todo list. It has a
unique task ID, which is different from the CKAN/DCOR dataset ID.
DCOR-Aid can keep track of dataset IDs and task IDs to avoid
uploading the same dataset twice.

Once a task is imported in DCOR-Aid, it becomes an
:class:`dcoraid.upload.UploadJob`. An upload job
can be converted back to a task using hte :func:`save_task`
method. The new task now automatically has a dataset ID
(given to it by CKAN/DCOR).
"""
import json
import pathlib

from ..api import dataset_create

from .job import UploadJob


class LocalTaskResourcesNotFoundError(FileNotFoundError):
    def __init__(self, missing_resources, *args, **kwargs):
        self.missing_resources = missing_resources
        super(LocalTaskResourcesNotFoundError, self).__init__(*args, **kwargs)


class PersistentTaskDatasetIDDict:
    def __init__(self, path):
        """A file-based dictionary with task_id as keys

        Features:

        - Not possible to override any keys with other values
        - allowed key characters are
          "0123456789-_abcdefghijklmnopqrstuvwxyz"
        - New keys are appended to the end of the file
        - No file locking (make sure that only one thread
          appends to the file)
        - Dictionary is loaded into memory on init, then for
          each new entry, the internal dictionary is updated
          and a new line is appended to the file on disk
        """
        self._path = pathlib.Path(path)
        self._path.touch(exist_ok=True)
        self._dict = {}
        # load the dictionary
        with self._path.open() as fd:
            lines = fd.readlines()
            for line in lines:
                task_id, dataset_id = line.strip().split()
                self._dict[task_id] = dataset_id

    def __contains__(self, task_id):
        return self._dict.__contains__(task_id)

    def __setitem__(self, task_id, dataset_id):
        if task_id in self._dict:
            if self[task_id] != dataset_id:
                raise ValueError("Cannot override entries in persistent dict!")
            else:
                # everything ok
                pass  # this line is covered, even though coverage disagrees
        else:
            valid_ch = "0123456789-_abcdefghijklmnopqrstuvwxyz"
            task_id_check = "".join([ch for ch in task_id if ch in valid_ch])
            if task_id_check != task_id:
                raise ValueError("task IDs may only contain numbers, "
                                 "lower-case characters, and '-_'."
                                 f"Got '{task_id}'!")
            # append to end of file
            with self._path.open("a") as fd:
                fd.write(f"{task_id} {dataset_id}\n")
            # append to dict
            self._dict[task_id] = dataset_id

    def __getitem__(self, task_id):
        return self._dict[task_id]

    def get(self, task_id, default=None):
        return self._dict.get(task_id, default)


def save_task(upload_job, path, dataset_dict=None):
    """Save an upload job to a JSON file

    Parameters
    ----------
    upload_job: dcoraid.upload.job.UploadJob
        Upload job from which to create a snapshot
    path: str or pathlib.Path
        Output path
    dataset_dict: dict
        An optional dataset dictionary. This dictionary is
        purely informative; it only contains redundant
        information (which is stored on the DCOR server).
    """
    path = pathlib.Path(path)
    uj_state = {"upload_job": upload_job.__getstate__()}
    if dataset_dict:
        uj_state["dataset_dict"] = dataset_dict
    with path.open("w") as fd:
        json.dump(uj_state, fd,
                  ensure_ascii=False,
                  indent=2,
                  sort_keys=True,
                  )


def load_task(path, api, dataset_kwargs=None, map_task_to_dataset_id=None,
              update_dataset_id=False, cache_dir=None):
    """Open a task file and load it into an UploadJob

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the JSON-encoded task file
    api: dcoraid.api.CKANAPI
        The CKAN/DCOR API instance used for the upload
    dataset_kwargs: dict
        Any additional dataset kwargs with which the "dataset_dict"
        dictionary in the task file should be updated. Note that
        it only make sense to change things like "owner_org" or
        "private" - you should e.g. never edit resource information
        (which will fail in an undefined manner). Note that if the
        dataset already exists (dataset_id) is defined, then the
        entries in this dictionary have no effect. They only have
        an effect when the dataset has to be created.
    map_task_to_dataset_id: dict or PersistentTaskDatasetIDDict
        Dictionary-like object that maps previously encountered
        task IDs to CKAN/DCOR dataset IDs. This is only used
        if there is no dataset ID in the task file. Use this if
        you want to avoid uploading duplicate datasets. If task_id
        is given in the task file, then map_task_to_dataset_id will
        be updated with the dataset_id.
    update_dataset_id: bool
        If True, update the task file with the dataset identifier
        assigned to by the CKAN/DCOR server.
    cache_dir: str or pathlib.Path
        Cache directory for storing compressed .rtdc files;
        if not supplied, a temporary directory is created

    Returns
    -------
    upload_job: dcoraid.upload.job.UploadJob
        The corresponding upload job
    """
    path = pathlib.Path(path)
    if map_task_to_dataset_id is None:
        # just set to empty dict so the code below may remain simple
        map_task_to_dataset_id = {}

    with path.open() as fd:
        task_dict = json.load(fd)

    if "dataset_dict" in task_dict:
        # separate "dataset_dict" from the task file
        # (e.g. the dataset_id and other things might be in here)
        dataset_dict = task_dict["dataset_dict"]
    else:
        dataset_dict = {}

    if dataset_kwargs is not None:
        dataset_dict.update(dataset_kwargs)

    uj_state = task_dict["upload_job"]

    # make sure the paths exist (if not, try with name relative to task path)
    missing_resources = []
    for ii in range(len(uj_state["resource_paths"])):
        pi = pathlib.Path(uj_state["resource_paths"][ii])
        pi_alt = pathlib.Path(path).parent / pi.name
        if pi.exists():
            pass
        elif pi_alt.exists():
            # replace with path relative to task file
            uj_state["resource_paths"][ii] = str(pi_alt)
        else:
            missing_resources.append(pi)

    if missing_resources:
        resstr = ", ".join([str(pp) for pp in missing_resources])
        raise LocalTaskResourcesNotFoundError(
            missing_resources,
            f"Task '{path}' is missing local resources files: {resstr}")

    num_res = len(uj_state["resource_paths"])
    if (uj_state["resource_names"]
            and len(uj_state["resource_names"]) != num_res):
        raise ValueError("Number of resource paths does not match number "
                         f"of resource names in '{path}'!")
    if uj_state["resource_supplements"]:
        if len(uj_state["resource_supplements"]) != num_res:
            raise ValueError("Number of resource paths does not match number "
                             f"of resource supplements in '{path}'!")
        for ii, pp in enumerate(uj_state["resource_paths"]):
            if (not str(pp).endswith(".rtdc")
                    and uj_state["resource_supplements"][ii]):
                raise ValueError("Resource supplements must be empty for "
                                 f"non-RT-DC datasets in '{path}'!")

    # determine the dataset id...
    # ...from the upload job state
    id_u = uj_state.get("dataset_id")
    # ...from the dataset_dict
    id_d = dataset_dict.get("id")
    # ...from the shelf using the task_id
    task_id = uj_state.get("task_id")
    if task_id is None:
        id_m = None
        if id_u is None and id_d is None:
            # This is an important point for automation. We can't just
            # have people running around not specifying any task_ids.
            raise ValueError(
                f"Neither task_id nor dataset_id are specified in '{path}'! "
                "Please update the task file or pass the dataset_id via "
                "the dataset_kwargs argument.")
    else:
        id_m = map_task_to_dataset_id.get(task_id)

    # use the ids to determine what to do next
    ids = [dd for dd in [id_u, id_d, id_m] if dd is not None]
    if len(set(ids)) == 0:
        # We have no dataset_id, so we create a dataset first
        ddict = dataset_create(
            dataset_dict=dataset_dict,
            api=api,
            activate=False)
        dataset_id = ddict["id"]
    elif len(set(ids)) == 1:
        # We have a unique dataset_id
        dataset_id = ids[0]
    else:
        # Something went wrong when creating the task file or the
        # user is insane.
        raise ValueError(
            "There are ambiguities regarding the dataset_id! I got the "
            + "following IDs: "
            + (f"from upload job state: {id_u}; " if id_u is not None else "")
            + (f"from dataset dict: {id_d}; " if id_d is not None else "")
            + (f"from ID dict: {id_m}; " if id_m is not None else "")
            + "Please check your input data!"
        )

    # Finally, set the dataset ID
    uj_state["dataset_id"] = dataset_id
    dataset_dict["id"] = dataset_id

    # Proceed with instantiation of UploadJob
    uj = UploadJob.from_upload_job_state(uj_state, api=api,
                                         cache_dir=cache_dir)

    if task_id is not None:
        # Also update the ID dictionary
        # This part here should come after instantiation of UploadJob,
        # because that just might fail if the specified dataset_id is invalid
        # (and we don't want invalid IDs in our dictionary).
        map_task_to_dataset_id[task_id] = dataset_id

    if update_dataset_id:
        # If everything went well so far, then we can safely update the
        # original json file with the new dictionary that contains
        # the dataset id.
        save_task(uj, path, dataset_dict=dataset_dict)

    return uj


def task_has_circle(path):
    """Check whether a circle (owner_org) is specified in a task file"""
    path = pathlib.Path(path)
    with path.open() as fd:
        task_dict = json.load(fd)
    has_org = ("owner_org" in task_dict.get("dataset_dict", {})
               and task_dict["dataset_dict"]["owner_org"])
    return has_org



from functools import lru_cache


class DBExtract:
    def __init__(self, datasets=None):
        """User-convenient access to dataset search results

        Parameters
        ----------
        datasets: list
            List of CKAN package dictionaries
        """
        self._circles = None
        self._collections = None
        self._dataset_name_index = None

        self.registry = {}
        self.datasets = []
        if datasets:
            self.add_datasets(datasets)

    def __add__(self, other):
        return DBExtract(self.datasets + other.datasets)

    def __iadd__(self, other):
        self.add_datasets(other.datasets)
        return self

    def __len__(self):
        return len(self.datasets)

    def __iter__(self):
        return iter(self.datasets)

    def add_datasets(self, datasets):
        for dd in datasets:
            name = dd["name"]
            if name not in self.registry:  # datasets must only be added once
                self.registry[name] = dd
                self.datasets.append(dd)

    @property
    @lru_cache(maxsize=1)
    def circles(self):
        if not self._circles:
            circ_list = []
            circ_names = []
            for dd in self.datasets:
                name = dd["organization"]["name"]
                if name not in circ_names:
                    circ_list.append(dd["organization"])
                    circ_names.append(name)
            self._circles = sorted(
                circ_list, key=lambda x: x.get("title") or x["name"])
        return self._circles

    @property
    @lru_cache(maxsize=1)
    def collections(self):
        if not self._collections:
            coll_list = []
            coll_names = []
            for dd in self.datasets:
                for gg in dd["groups"]:
                    name = gg["name"]
                    if name not in coll_names:
                        coll_list.append(gg)
                        coll_names.append(name)
            self._collections = sorted(
                coll_list, key=lambda x: x.get("title") or x["name"])
        return self._collections

    def get_dataset_dict(self, dataset_name):
        return self.registry[dataset_name]



# flake8: noqa: F401
from .ckan_api import CKANAPI
from .dataset import (dataset_create, dataset_activate, dataset_draft_remove,
                      dataset_draft_remove_all)
from .dataset import resource_add, resource_exists, resource_sha256_sums
from .errors import (APIError, APIBadGatewayError, APIConflictError,
                     APIAuthorizationError, APIKeyError, APINotFoundError,
                     APIGatewayTimeoutError)



class APIError(BaseException):
    pass


class APIAuthorizationError(APIError):
    pass


class APIBadGatewayError(APIError):
    pass


class APIConflictError(APIError):
    pass


class APIGatewayTimeoutError(APIError):
    pass


class APIKeyError(APIError):
    pass


class APINotFoundError(APIError):
    pass



import os.path as os_path
import pathlib
import shutil
import tempfile
import time

from PyQt5 import QtCore
from PyQt5.QtCore import QStandardPaths
import pytest

from dcoraid.api import APIConflictError

import common


TMPDIR = tempfile.mkdtemp(prefix=time.strftime(
    "dcoraid_test_%H.%M_"))

pytest_plugins = ["pytest-qt"]


def pytest_configure(config):
    """This is ran before all tests"""
    # disable update checking
    QtCore.QCoreApplication.setOrganizationName("DCOR")
    QtCore.QCoreApplication.setOrganizationDomain("dcor.mpl.mpg.de")
    QtCore.QCoreApplication.setApplicationName("dcoraid")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setIniCodec("utf-8")
    settings.value("user scenario", "dcor-dev")
    settings.setValue("auth/server", "dcor-dev.mpl.mpg.de")
    settings.setValue("auth/api key", common.get_api_key())
    settings.setValue("debug/without timers", 1)
    settings.sync()
    # remove persistent upload jobs
    shelf_path = os_path.join(
        QStandardPaths.writableLocation(
            QStandardPaths.AppLocalDataLocation),
        "persistent_upload_jobs")
    shutil.rmtree(shelf_path, ignore_errors=True)
    # remove persistent upload id dict
    path_id_dict = os_path.join(
        QStandardPaths.writableLocation(
            QStandardPaths.AppLocalDataLocation),
        "map_task_to_dataset_id.txt")
    path_id_dict = pathlib.Path(path_id_dict)
    if path_id_dict.exists():
        path_id_dict.unlink()
    # set global temp directory
    tempfile.tempdir = TMPDIR


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """
    settings = QtCore.QSettings()
    settings.setIniCodec("utf-8")
    settings.remove("debug/without timers")
    settings.sync()
    # clear global temp directory
    shutil.rmtree(TMPDIR, ignore_errors=True)


def pytest_sessionstart(session):
    """
    Called after the Session object has been created and
    before performing collection and entering the run test loop.
    """
    api = common.get_api()
    api.get("group_show", id=common.COLLECTION)
    try:
        api.post("group_create", {"name": common.COLLECTION})
    except APIConflictError:
        pass
    try:
        api.post("organization_create", {"name": common.CIRCLE})
    except APIConflictError:
        pass
    user_dict = api.get("user_show", id=common.USER)
    user_dict["fullname"] = common.USER_NAME
    api.post("user_update", user_dict)


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """Writes report to failures file

    We need this to check whether the tests passed or failed
    du to the threading issue we have.
    https://github.com/DCOR-dev/DCOR-Aid/issues/14
    """
    # https://docs.pytest.org/en/stable/example/simple.html
    # #post-process-test-reports-failures
    # execute all other hooks to obtain the report object
    outcome = yield
    rep = outcome.get_result()

    # we only look at actual failing test calls, not setup/teardown
    if rep.when == "call" and rep.failed:
        mode = "a" if os_path.exists("pytest-failures.txt") else "w"
        with open("pytest-failures.txt", mode) as f:
            # let's also access a fixture for the fun of it
            if "tmpdir" in item.fixturenames:
                extra = " ({})".format(item.funcargs["tmpdir"])
            else:
                extra = ""
            f.write(rep.nodeid + extra + "\n")



from functools import partial
import webbrowser

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import Qt

from ..api import get_ckan_api
from ..tools import ShowWaitCursor

from . import filter_base


class FilterCircles(filter_base.FilterBase):
    def __init__(self, *args, **kwargs):
        super(FilterCircles, self).__init__(*args, **kwargs)
        self.label.setText("Circles")
        self.lineEdit.setPlaceholderText("filter names...")
        self.checkBox.setVisible(False)
        self.label_info.setVisible(False)

    def get_entry_actions(self, row, entry):
        api = get_ckan_api()
        url = f"{api.server}/organization/{entry['name']}"
        actions = [
            {"icon": "eye",
             "tooltip": f"view circle {entry['name']} online",
             "function": partial(webbrowser.open, url)}
        ]
        return actions


class FilterCollections(filter_base.FilterBase):
    def __init__(self, *args, **kwargs):
        super(FilterCollections, self).__init__(*args, **kwargs)
        self.label.setText("Collections")
        self.lineEdit.setPlaceholderText("filter names...")
        self.checkBox.setVisible(False)
        self.label_info.setVisible(False)

    def get_entry_actions(self, row, entry):
        api = get_ckan_api()
        url = f"{api.server}/group/{entry['name']}"
        actions = [
            {"icon": "download",
             "tooltip": f"download collection {entry['name']}",
             "function": partial(self.download_collection, entry["name"])},
            {"icon": "eye",
             "tooltip": f"view collection {entry['name']} online",
             "function": partial(webbrowser.open, url)}
        ]
        return actions

    def download_collection(self, collection_name):
        with ShowWaitCursor():
            api = get_ckan_api()
            search_dict = api.get("package_search",
                                  fq=f"+groups:{collection_name}",
                                  include_private=True,
                                  rows=1000)
            num_datasets = search_dict["count"]
            if num_datasets >= 1000:
                raise NotImplementedError(
                    # We have to increase ckan.search.rows_max = 1000
                    # or (better) use the `start` parameter until we
                    # hit a number < 1000.
                    f"There are too many datasets in '{collection_name}'!")
            for ii, ds_dict in enumerate(search_dict["results"]):
                for res_dict in ds_dict["resources"]:
                    self.download_resource.emit(res_dict["id"])
                    QtWidgets.QApplication.processEvents(
                        QtCore.QEventLoop.AllEvents,
                        300)


class FilterDatasets(filter_base.FilterBase):
    def __init__(self, *args, **kwargs):
        super(FilterDatasets, self).__init__(*args, **kwargs)
        self.label.setText("Datasets")
        self.lineEdit.setPlaceholderText("filter titles...")
        self.checkBox.setVisible(False)
        self.label_info.setVisible(False)

    def get_entry_actions(self, row, entry):
        api = get_ckan_api()
        url = f"{api.server}/dataset/{entry['name']}"
        actions = [
            {"icon": "download",
             "tooltip": f"download dataset {entry['name']}",
             "function": partial(self.download_dataset, entry["id"])},
            {"icon": "eye",
             "tooltip": f"view dataset {entry['name']} online",
             "function": partial(webbrowser.open, url)},
        ]
        return actions

    @QtCore.pyqtSlot(str)
    def download_dataset(self, dataset_id):
        api = get_ckan_api()
        ds_dict = api.get("package_show", id=dataset_id)
        for res_dict in ds_dict["resources"]:
            self.download_resource.emit(res_dict["id"])
            QtWidgets.QApplication.processEvents(
                QtCore.QEventLoop.AllEvents,
                300)


class FilterResources(filter_base.FilterBase):
    def __init__(self, *args, **kwargs):
        super(FilterResources, self).__init__(*args, **kwargs)
        self.label.setText("Resources")
        self.lineEdit.setPlaceholderText("filter file names...")
        self.checkBox.setVisible(True)
        self.checkBox.setText(".rtdc only")
        self.checkBox.setChecked(True)
        self.tableWidget.setDragEnabled(True)
        self.label_info.setText("<i>Tip: You can drag and drop your selection "
                                "from the resources list to Shape-Out!</i>")

    def get_entry_actions(self, row, entry):
        api = get_ckan_api()
        url = f"{api.server}/dataset/{entry['package_id']}/" \
              + f"resource/{entry['id']}"
        actions = [
            {"icon": "download",
             "tooltip": f"download resource {entry['name']}",
             "function": partial(self.download_resource.emit, entry["id"])},
            {"icon": "eye",
             "tooltip": f"view resource {entry['name']} online",
             "function": partial(webbrowser.open, url)},
        ]
        return actions

    def set_entry_label(self, row, entry):
        super(FilterResources, self).set_entry_label(row, entry)
        # set additional data link
        item = self.tableWidget.item(row, 0)
        api = get_ckan_api()
        dcor_url = f"{api.server}/api/3/action/dcserv?id={entry['id']}"
        item.setData(Qt.UserRole + 1, dcor_url)



import copy

import pkg_resources

from PyQt5 import QtCore, QtWidgets, uic


class FilterChain(QtWidgets.QWidget):
    download_resource = QtCore.pyqtSignal(str)

    def __init__(self, *args, **kwargs):
        """Filter chain widget with multiple filter views
        """
        super(FilterChain, self).__init__(*args, **kwargs)
        QtWidgets.QMainWindow.__init__(self)
        path_ui = pkg_resources.resource_filename("dcoraid.gui.dbview",
                                                  "filter_chain.ui")
        uic.loadUi(path_ui, self)

        #: Current database extract
        #: (instance of :class:`dcoraid.dbmodel.core.DBExtract`)
        self.db_extract = None

        # Signals and slots
        # circle selection changed
        self.fw_circles.selection_changed.connect(self.on_select_circles)
        # collection selection changed
        self.fw_collections.selection_changed.connect(self.update_datasets)
        # dataset selection changed
        self.fw_datasets.selection_changed.connect(self.update_resources)
        # resource filtered by .rtdc
        self.fw_resources.checkBox.stateChanged.connect(self.update_resources)
        # request resource downloads
        self.fw_circles.download_resource.connect(self.download_resource)
        self.fw_collections.download_resource.connect(self.download_resource)
        self.fw_datasets.download_resource.connect(self.download_resource)
        self.fw_resources.download_resource.connect(self.download_resource)

    @property
    def selected_circles(self):
        """Circles currently selected"""
        circs = self.fw_circles.get_entry_identifiers(selected=True)
        if not circs:
            circs = self.fw_circles.get_entry_identifiers(selected=False)
        return circs

    @property
    def selected_collections(self):
        """Collections currently selected"""
        return self.fw_collections.get_entry_identifiers(selected=True)

    @property
    def selected_datasets(self):
        """Datasets currently selected"""
        return self.fw_datasets.get_entry_identifiers(selected=True)

    @QtCore.pyqtSlot()
    def on_select_circles(self):
        self.update_collections()
        self.update_datasets()

    def set_db_extract(self, db_extract):
        """Set the database model

        Parameters
        ----------
        db_extract: dcoraid.dbmodel.db_core.DBExtract
            Subclass of DBModel
        """
        self.db_extract = db_extract
        # reset all lists
        self.fw_circles.set_entries(
            copy.deepcopy(self.db_extract.circles))
        # collections
        self.update_collections()
        # datasets
        self.update_datasets()

    def update_collections(self):
        self.fw_collections.blockSignals(True)
        self.fw_collections.set_entries(
            copy.deepcopy(self.db_extract.collections))
        self.fw_collections.blockSignals(False)

    @QtCore.pyqtSlot()
    def update_datasets(self):
        circles = self.selected_circles
        collections = self.selected_collections
        dataset_items = []
        for ds in self.db_extract.datasets:
            if ds["organization"]["name"] not in circles:
                # dataset is not part of the circle
                continue
            if collections:
                ds_collections = [g.get("name") for g in ds.get("groups", {})]
                if not set(collections) & set(ds_collections):
                    # collections have been selected and the dataset is not
                    # part of any
                    continue
            dataset_items.append(ds)
        self.fw_datasets.blockSignals(True)
        self.fw_datasets.set_entries(dataset_items)
        self.fw_datasets.blockSignals(False)
        self.update_resources()

    @QtCore.pyqtSlot()
    def update_resources(self):
        rs_entries = []
        for dn in self.selected_datasets:
            ddict = self.db_extract.get_dataset_dict(dn)
            for rs in ddict["resources"]:
                if (self.fw_resources.checkBox.isChecked()
                    and ("mimetype" in rs
                         and rs["mimetype"] != "RT-DC")):
                    # Ignore non-RT-DC mimetypes
                    continue
                else:
                    rs_entries.append(rs)
        self.fw_resources.set_entries(rs_entries)



from functools import partial
import os
import os.path as os_path
import pkg_resources
import platform
import subprocess
import webbrowser

from PyQt5 import uic, QtCore, QtGui, QtWidgets
from PyQt5.QtCore import QStandardPaths

from ...download import DownloadQueue

from ..api import get_ckan_api


class DownloadWidget(QtWidgets.QWidget):
    download_finished = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        """Manage running downloads
        """
        super(DownloadWidget, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.download", "widget_download.ui")
        uic.loadUi(path_ui, self)

        # use a persistent shelf to be able to resume uploads on startup
        shelf_path = os_path.join(
            QStandardPaths.writableLocation(
                QStandardPaths.AppLocalDataLocation),
            "persistent_download_jobs")

        self.jobs = DownloadQueue(api=get_ckan_api(),
                                  path_persistent_job_list=shelf_path)
        self.widget_jobs.set_job_list(self.jobs)

    @QtCore.pyqtSlot(str)
    def download_resource(self, resource_id):
        dl_path = QStandardPaths.writableLocation(
                      QStandardPaths.DownloadLocation)
        self.widget_jobs.jobs.new_job(resource_id, dl_path)


class DownloadTableWidget(QtWidgets.QTableWidget):
    download_finished = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(DownloadTableWidget, self).__init__(*args, **kwargs)
        self.jobs = []  # Will become DownloadQueue with self.set_job_list

        settings = QtCore.QSettings()
        if bool(int(settings.value("debug/without timers", "0"))):
            self.timer = None
        else:
            self.timer = QtCore.QTimer()
            self.timer.timeout.connect(self.update_job_status)
            self.timer.start(1000)
        self._finished_downloads = []

    def set_job_list(self, jobs):
        """Set the current job list

        The job list can be a `list`, but it is actually
        a `DownloadQueue`.
        """
        # This is the actual initialization
        self.jobs = jobs

    def on_job_abort(self, resource_id):
        self.jobs.abort_job(resource_id)

    def on_job_delete(self, resource_id):
        self.jobs.remove_job(resource_id)
        self.clearContents()
        self.update_job_status()

    def on_download_finished(self, resource_id):
        """Triggers download_finished whenever a download is finished"""
        if resource_id not in self._finished_downloads:
            self._finished_downloads.append(resource_id)
            self.jobs.jobs_eternal.set_job_done(resource_id)
            self.download_finished.emit()

    @QtCore.pyqtSlot()
    def update_job_status(self):
        """Update UI with information from self.jobs (DownloadJobList)"""
        # disable updates
        self.setUpdatesEnabled(False)
        # make sure the length of the table is long enough
        self.setRowCount(len(self.jobs))
        self.setColumnCount(6)

        for row, job in enumerate(self.jobs):
            status = job.get_status()
            self.set_label_item(row, 0, job.resource_id[:5])
            try:
                title = get_download_title(job)
            except BaseException:
                # Probably a connection error
                title = "-- error getting dataset title --"
            self.set_label_item(row, 1, title)
            self.set_label_item(row, 2, status["state"])
            self.set_label_item(row, 3, job.get_progress_string())
            self.set_label_item(row, 4, job.get_rate_string())
            if status["state"] == "done":
                self.on_download_finished(job.resource_id)
            self.set_actions_item(row, 5, job)

        # spacing (did not work in __init__)
        header = self.horizontalHeader()
        header.setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)
        # enable updates again
        self.setUpdatesEnabled(True)

    def set_label_item(self, row, col, label):
        """Get/Create a Qlabel at the specified position

        User has to make sure that row and column count are set
        """
        label = "{}".format(label)
        item = self.item(row, col)
        if item is None:
            item = QtWidgets.QTableWidgetItem(label)
            item.setFlags(QtCore.Qt.ItemIsEnabled)
            self.setItem(row, col, item)
        else:
            if item.text() != label:
                item.setText(label)

    def set_actions_item(self, row, col, job):
        """Set/Create a TableCellActions widget in the table

        Refreshes the widget and also connects signals.
        """
        widact = self.cellWidget(row, col)
        if widact is None:
            widact = QtWidgets.QWidget(self)
            horz_layout = QtWidgets.QHBoxLayout(widact)
            horz_layout.setContentsMargins(2, 0, 2, 0)

            spacer = QtWidgets.QSpacerItem(0, 0,
                                           QtWidgets.QSizePolicy.Expanding,
                                           QtWidgets.QSizePolicy.Minimum)
            horz_layout.addItem(spacer)

            res_dict = job.get_resource_dict()
            ds_dict = job.get_dataset_dict()
            dl_path = job.path
            if not dl_path.is_dir():
                dl_path = dl_path.parent
            actions = [
                {"icon": "eye",
                 "tooltip": f"view dataset {ds_dict['name']} online",
                 "function": partial(
                     webbrowser.open,
                     f"{job.api.server}/dataset/{ds_dict['id']}")
                 },
                {"icon": "folder",
                 "tooltip": "open local download directory",
                 "function": partial(open_file, str(dl_path))
                 },
                {"icon": "trash",
                 "tooltip": f"abort download {res_dict['name']}",
                 "function": partial(self.on_job_delete, job.resource_id)
                 },
            ]
            for action in actions:
                tbact = QtWidgets.QToolButton(widact)
                icon = QtGui.QIcon.fromTheme(action["icon"])
                tbact.setIcon(icon)
                tbact.setToolTip(action["tooltip"])
                tbact.clicked.connect(action["function"])
                horz_layout.addWidget(tbact)
            self.setCellWidget(row, col, widact)


def get_download_title(job):
    res_dict = job.get_resource_dict()
    ds_dict = job.get_dataset_dict()
    title = ds_dict.get("title")
    if not title:
        title = ds_dict.get("name")
    return f"{res_dict['name']} [{title}]"


def open_file(path):
    if platform.system() == "Windows":
        os.startfile(path)
    elif platform.system() == "Darwin":
        subprocess.Popen(["open", path])
    else:
        subprocess.Popen(["xdg-open", path])



import pathlib
import warnings

from ..worker import Daemon

from .job import DownloadJob
from .task import load_task, save_task


class DCORAidQueueWarning(UserWarning):
    pass


class PersistentDownloadJobList:
    def __init__(self, path):
        """A file-system and JSON-based persistent DownloadJob list"""
        self.path = pathlib.Path(path)
        self.path_queued = self.path / "queued"
        self.path_queued.mkdir(parents=True, exist_ok=True)

    def __contains__(self, item):
        if isinstance(item, DownloadJob):
            resource_id = item.resource_id
        else:
            resource_id = item
        return self.job_exists(resource_id)

    @property
    def num_queued(self):
        """Return number of queued tasks"""
        return len(list(self.path_queued.glob("*.json")))

    def get_queued_resource_ids(self):
        """Return list of DCOR resource IDs corresponding to queued jobs"""
        return sorted([pp.stem for pp in self.path_queued.glob("*.json")])

    def is_job_queued(self, resource_id):
        jp = self.path_queued / (resource_id + ".json")
        return jp.exists()

    def immortalize_job(self, download_job):
        """Put this job in the persistent queue list"""
        pout = self.path_queued / (download_job.resource_id + ".json")
        if self.is_job_queued(download_job.resource_id):
            raise FileExistsError(f"The job '{download_job.resource_id}' is "
                                  f"already present at '{pout}'!")
        save_task(download_job=download_job, path=pout)

    def job_exists(self, resource_id):
        return self.is_job_queued(resource_id)

    def obliterate_job(self, resource_id):
        """Remove a job from the persistent queue list"""
        pdel = self.path_queued / (resource_id + ".json")
        pdel.unlink()

    def set_job_done(self, resource_id):
        """Remove a job from the persistent queue list"""
        self.obliterate_job(resource_id)

    def summon_job(self, resource_id, api, cache_dir=None):
        """Instantiate job from the persistent queue list"""
        pin = self.path_queued / (resource_id + ".json")
        download_job = load_task(path=pin, api=api)
        assert download_job.resource_id == resource_id
        return download_job


class DownloadQueue:
    def __init__(self, api, path_persistent_job_list=None):
        """Manager for running multiple downloadJobs in sequence

        Parameters
        ----------
        api: dcoraid.api.CKANAPI
            The CKAN/DCOR API instance used for the downloads
        path_persistent_job_list: str or pathlib.Path
            Path to a directory for storing downloadJobs in a
            persistent manner across restarts.
        """
        self.api = api.copy()
        if not api.api_key:
            warnings.warn("No API key is set! download may not work!")
        self.jobs = []
        if path_persistent_job_list is not None:
            self.jobs_eternal = PersistentDownloadJobList(
                path_persistent_job_list)
            # add any previously queued jobs
            for resource_id in self.jobs_eternal.get_queued_resource_ids():
                dj = self.jobs_eternal.summon_job(resource_id, api=self.api)
                self.jobs.append(dj)
        else:
            self.jobs_eternal = None
        self.daemon_download = DownloadDaemon(self.jobs)
        self.daemon_verify = VerifyDaemon(self.jobs)

    def __contains__(self, download_job):
        return download_job in self.jobs

    def __getitem__(self, index):
        return self.jobs[index]

    def __len__(self):
        return len(self.jobs)

    def abort_job(self, resource_id):
        """Abort a running job but don't remove it from the queue"""
        job = self.get_job(resource_id)
        if job.state == "transfer":
            job.set_state("abort")
            # https://github.com/requests/toolbelt/issues/297
            self.daemon_download.terminate()
            self.daemon_download = DownloadDaemon(self.jobs)

    def add_job(self, download_job):
        """Add a DownloadJob to the queue"""
        if self.jobs_eternal is not None:
            if download_job in self.jobs_eternal:
                self.jobs_eternal.obliterate_job(download_job.resource_id)
            # Add to eternal jobs for persistence
            self.jobs_eternal.immortalize_job(download_job)
        try:
            self.get_job(download_job.resource_id)
        except KeyError:
            self.jobs.append(download_job)

    def get_job(self, resource_id):
        """Return the queued DownloadJob belonging to the resource ID"""
        for job in self.jobs:
            if job.resource_id == resource_id:
                return job
        else:
            raise KeyError("Job '{}' not found!".format(resource_id))

    def get_status(self, resource_id):
        """Return the status of an downloadJob"""
        self.get_job(resource_id).get_status()

    def new_job(self, resource_id, download_path):
        """Create an downloadJob and add it to the download queue

        Parameters
        ----------
        resource_id: str
            The CKAN/DCOR dataset ID
        download_path: str or pathlib.Path
            Download path

        Returns
        -------
        download_job: dcoraid.download.job.DownloadJob
            The download job that was appended to the download queue
        """
        download_job = DownloadJob(
            api=self.api,
            resource_id=resource_id,
            download_path=download_path,
        )
        self.add_job(download_job)
        return download_job

    def remove_job(self, resource_id):
        """Remove a DownloadJob from the queue and perform cleanup

        Running jobs are aborted before they are removed.
        """
        self.abort_job(resource_id)
        for ii, job in enumerate(list(self.jobs)):
            if job.resource_id == resource_id:
                self.jobs.pop(ii)
                # cleanup temp files
                try:
                    job.path_temp.unlink()
                except BaseException:
                    pass
            elif job.state == "abort":
                job.set_state("init")
        # also remove from eternal jobs
        if (self.jobs_eternal is not None
                and self.jobs_eternal.is_job_queued(resource_id)):
            self.jobs_eternal.obliterate_job(resource_id)


class DownloadDaemon(Daemon):
    def __init__(self, jobs):
        """Download daemon"""
        super(DownloadDaemon, self).__init__(
            jobs,
            job_trigger_state="init",
            job_function_name="task_download_resource")


class VerifyDaemon(Daemon):
    def __init__(self, jobs):
        """Verify daemon"""
        super(VerifyDaemon, self).__init__(
            jobs,
            job_trigger_state="downloaded",
            job_function_name="task_verify_resource")



import shutil

import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets

from ...api import dataset_draft_remove_all

from ..api import get_ckan_api
from ..main import DCORAid
from ..tools import ShowWaitCursor


class MaintenanceWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs):
        """Maintenance tasks
        """
        super(MaintenanceWidget, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.maintenance", "widget_maintenance.ui")
        uic.loadUi(path_ui, self)

        self.toolButton_cache.clicked.connect(self.on_clear_cache)
        self.toolButton_drafts.clicked.connect(self.on_remove_drafts)

    @staticmethod
    def find_main_window():
        # Global function to find the (open) QMainWindow in application
        app = QtWidgets.QApplication.instance()
        for widget in app.topLevelWidgets():
            if isinstance(widget, DCORAid):
                return widget

    @QtCore.pyqtSlot()
    def on_clear_cache(self):
        """Clear local upload cache"""
        mw = self.find_main_window()
        queue = mw.panel_upload.jobs
        dirs = queue.find_zombie_caches()
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        if dirs:
            [shutil.rmtree(pp, ignore_errors=True) for pp in dirs]
            details = [f"- {pp}" for pp in dirs]
            msg.setText(f"Directories removed: {len(dirs)}")
            msg.setDetailedText("\n".join(details))
            msg.setWindowTitle("Success")
        else:
            msg.setText("No zombie cache data found.")
            msg.setWindowTitle("Nothing to do")
        msg.exec_()

    @QtCore.pyqtSlot()
    def on_remove_drafts(self):
        with ShowWaitCursor():
            data = dataset_draft_remove_all(api=get_ckan_api())
        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Information)
        if len(data):
            details = [f"{d['title']} ({d['name']})" for d in data]
            msg.setText(f"Drafts removed: {len(data)}")
            msg.setDetailedText("\n".join(details))
            msg.setWindowTitle("Success")
        else:
            msg.setText("No drafts found.")
            msg.setWindowTitle("Nothing to do")
        msg.exec_()



"""Convenience wrappers for commonly used API calls"""

import copy
import pathlib

from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor


def dataset_activate(dataset_id, api):
    """Change the state of a dataset to "active"

    In the DCOR workflow, datasets are created as drafts and
    resources are added to the drafts. After that, the
    datasets are activated (and no resources can be added
    anymore).

    Parameters
    ----------
    dataset_id: str
        CKAN ID of the dataset
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    """
    revise_dict = {
        "match": {"id": dataset_id},
        "update": {"state": "active"}}
    api.post("package_revise", revise_dict)


def dataset_create(dataset_dict, api, resources=None,
                   create_circle=False, activate=False):
    """Create a draft dataset

    Parameters
    ----------
    dataset_dict: dict
        CKAN dataset dictionary
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    resources: list of str or list of pathlib.Path
        Paths to dataset resources
    create_circle: bool
        Create the circle if it does not already exist
    activate: bool
        If True, then the dataset state is changed to "active"
        after uploading of the resources is complete. For DCOR,
        this implies that no other resources can be added to the
        dataset.
    """
    if resources is None:
        resources = []
    if create_circle:
        circles = [c["name"] for c in api.get("organization_list_for_user")]
        if dataset_dict["owner_org"] not in circles:
            # Create the circle before creating the dataset
            api.post("organization_create",
                     data={"name": dataset_dict["owner_org"]})
    dataset_dict = copy.deepcopy(dataset_dict)
    dataset_dict["state"] = "draft"
    data = api.post("package_create", dataset_dict)
    if resources:
        for res in resources:
            resource_add(dataset_id=data["id"],
                         path=res,
                         api=api)
    if activate:
        dataset_activate(dataset_id=data["id"], api=api)
        data["state"] = "active"
    return data


def dataset_draft_remove(dataset_id, api):
    """Remove a draft dataset

    Use this for deleting datasets that you created but which were
    not yet activated. The dataset is deleted and purged.

    Parameters
    ----------
    dataset_id: str
        CKAN ID of the dataset to which the resource is added
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    """
    api.post("package_delete", {"id": dataset_id})
    api.post("dataset_purge", {"id": dataset_id})


def dataset_draft_remove_all(api):
    """Remove all draft datasets

    Find and delete all draft datasets for a user. The user
    ID is inferred from the API key.

    Parameters
    ----------
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    """
    user_dict = api.get_user_dict()
    data = api.get(
        "package_search",
        q="*:*",
        include_drafts=True,
        include_private=True,
        fq="creator_user_id:{} AND state:draft".format(user_dict["id"]),
        rows=1000)
    for dd in data["results"]:
        assert dd["state"] == "draft"
        dataset_draft_remove(dd["id"], api=api)
    return data["results"]


def resource_add(dataset_id, path, api, resource_name=None,
                 resource_dict=None, exist_ok=False, monitor_callback=None):
    """Add a resource to a dataset

    Parameters
    ----------
    dataset_id: str
        CKAN ID of the dataset to which the resource is added
    path: str or pathlib.Path
        Path to the resource
    api: dcoraid.api.CKANAPI
        API instance with server, api_key (and optional certificate)
    resource_name: str
        Alternate resource name, if not set `path.name` is used
    resource_dict: dict
        Dictionary of resource meta data (used for supplementary
        resource schemas "sp:section:key")
    exist_ok: bool
        If the uploaded resource already exists, do not re-upload
        it, only update the resource_dict.
    monitor_callback: None or callable
        This callable is used to monitor the upload progress. It must
        accept one argument: a
        `requests_toolbelt.MultipartEncoderMonitor` object.

    See Also
    --------
    dcoraid.upload.queue.UploadJob
        An implementation of an upload job that monitors progress.
    """
    if resource_dict is None:
        resource_dict = {}
    path = pathlib.Path(path)
    if resource_name is None:
        resource_name = path.name

    pkg_dict = api.get("package_show", id=dataset_id)
    res_names = [r["name"] for r in pkg_dict["resources"]]
    if resource_name in res_names and exist_ok:
        # Resource already uploaded and user does not wish to re-upload
        data = pkg_dict["resources"][res_names.index(resource_name)]
    else:
        # Attempt upload
        e = MultipartEncoder(fields={
            'package_id': dataset_id,
            'name': resource_name,
            'upload': (resource_name, path.open('rb'))})
        m = MultipartEncoderMonitor(e, monitor_callback)
        # perform upload
        data = api.post("resource_create",
                        data=m,
                        dump_json=False,
                        headers={"Content-Type": m.content_type})
    if resource_dict:
        # add resource_dict
        revise_dict = {
            "match": {"id": dataset_id},
            "update__resources__{}".format(data["id"]): resource_dict}
        api.post("package_revise", revise_dict)


def resource_exists(dataset_id, resource_name, api, resource_dict=None):
    """Check whether a resource exists in a dataset

    Parameters
    ----------
    dataset_id: str
        UUID of the dataset
    resource_name: str
        name of the resource
    api: dcoraid.api.CKANAPI
        API instance
    resource_dict: dict
        resource dictionary to check against (optional); If this is given,
        then this method returns False if there are discrepancies in the
        resource schema supplements (even if the resource exists).
    """
    if resource_dict is None:
        resource_dict = {}
    pkg_dict = api.get("package_show", id=dataset_id)
    for resource in pkg_dict["resources"]:
        if resource["name"] == resource_name:
            # check that the resource dict matches
            for key in resource_dict:
                if key not in resource or resource[key] != resource_dict[key]:
                    # Either the resource schema supplement is missing
                    # or it is wrong.
                    return False
            return True
    else:
        return False


def resource_sha256_sums(dataset_id, api):
    """Return a dictionary of resources with the SHA256 sums as values"""
    pkg_dict = api.get("package_show", id=dataset_id)
    sha256dict = {}
    for resource in pkg_dict["resources"]:
        sha256dict[resource["name"]] = resource.get("sha256", None)
    return sha256dict



import json
import pathlib

from .job import DownloadJob


def save_task(download_job, path):
    """Save a download job to a JSON file

    Parameters
    ----------
    download_job: dcoraid.download.job.DownloadJob
        Download job from which to create a snapshot
    path: str or pathlib.Path
        Output path
    """
    path = pathlib.Path(path)
    dj_state = {"download_job": download_job.__getstate__()}
    with path.open("w") as fd:
        json.dump(dj_state, fd,
                  ensure_ascii=False,
                  indent=2,
                  sort_keys=True,
                  )


def load_task(path, api):
    """Open a task file and load it into a DownloadJob

    Parameters
    ----------
    path: str or pathlib.Path
        Path to the JSON-encoded task file
    api: dcoraid.api.CKANAPI
        The CKAN/DCOR API instance used for the download
    """
    path = pathlib.Path(path)
    state = json.loads(path.read_text())
    return DownloadJob.from_download_job_state(state["download_job"], api=api)



import pathlib
import tempfile

import pytest

from dcoraid.download import DownloadJob, DownloadQueue
from dcoraid.download.queue import PersistentDownloadJobList

import common


data_path = pathlib.Path(__file__).parent / "data"
dpath = data_path / "calibration_beads_47.rtdc"


def test_queue_basic_functionalities():
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    ds_dict = common.make_dataset_for_download()
    joblist = DownloadQueue(api=api)
    resource_id = ds_dict["resources"][0]["id"]
    dj = joblist.new_job(resource_id=resource_id,
                         download_path=td)

    assert dj.resource_id == resource_id
    assert dj in joblist

    samejob = joblist.get_job(resource_id)
    assert samejob is dj

    with pytest.raises(KeyError, match="Job 'peter' not found!"):
        joblist.get_job("peter")

    samejob2 = joblist[0]
    assert samejob2 is dj


def test_queue_remove_job():
    """Remove a job from the queue and from the persistent list"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_dj_list_"))
    pdjl_path = td / "joblistdir"
    # create some metadata
    ds_dict = common.make_dataset_for_download()
    joblist = DownloadQueue(api=api,
                            path_persistent_job_list=pdjl_path)
    # disable all daemons, so no downloading happens
    joblist.daemon_download.join()
    joblist.daemon_verify.join()
    resource_id = ds_dict["resources"][0]["id"]
    dj = joblist.new_job(resource_id=resource_id,
                         download_path=td)
    assert dj.state == "init"
    joblist.remove_job(resource_id)
    assert dj not in joblist
    assert not joblist.jobs_eternal.job_exists(resource_id)

    # adding it again should work
    dj2 = joblist.new_job(resource_id=resource_id,
                          download_path=td)
    assert dj2 in joblist
    assert dj2.__getstate__() == dj.__getstate__()


def test_persistent_download_joblist_basic():
    """basic job tests"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_dj_list_"))
    pdjl_path = td / "joblistdir"
    # create some metadata
    ds_dict = common.make_dataset_for_download(seed="basic")
    joblist = DownloadQueue(api=api)
    resource_id = ds_dict["resources"][0]["id"]
    dj = joblist.new_job(resource_id=resource_id,
                         download_path=td)
    pdjl = PersistentDownloadJobList(pdjl_path)

    # add a job
    pdjl.immortalize_job(dj)
    assert dj in pdjl
    assert dj.resource_id in pdjl

    # find that job
    dj_same = pdjl.summon_job(resource_id, api=api)
    assert dj_same is not dj, "not same instance"
    assert dj_same.__getstate__() == dj.__getstate__(), "same data"
    ids = pdjl.get_queued_resource_ids()
    assert dj.resource_id in ids

    # remove a job
    assert pdjl.job_exists(dj.resource_id)
    assert pdjl.is_job_queued(dj.resource_id)
    pdjl.obliterate_job(dj.resource_id)
    assert dj not in pdjl
    assert not pdjl.job_exists(dj.resource_id)


def test_persistent_download_joblist_job_added_in_queue():
    """Test whether queuing works"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_dj_list_"))
    pdjl_path = td / "joblistdir"
    # create some metadata
    ds_dict = common.make_dataset_for_download(seed="job_added_in_queue")
    joblist = DownloadQueue(api=api)
    resource_id = ds_dict["resources"][0]["id"]
    dj = joblist.new_job(resource_id=resource_id,
                         download_path=td)
    pdjl = PersistentDownloadJobList(pdjl_path)

    uq = DownloadQueue(api=api, path_persistent_job_list=pdjl_path)
    uq.daemon_download.join()
    uq.daemon_verify.join()

    assert pdjl.num_queued == 0
    uq.add_job(dj)
    assert pdjl.num_queued == 1


def test_persistent_download_joblist_error_exists():
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_dj_list_"))
    pdjl_path = td / "joblistdir"
    ds_dict = common.make_dataset_for_download()
    pdjl = PersistentDownloadJobList(pdjl_path)
    joblist = DownloadQueue(api=api)
    resource_id = ds_dict["resources"][0]["id"]
    dj = joblist.new_job(resource_id=resource_id,
                         download_path=td)
    pdjl.immortalize_job(dj)
    with pytest.raises(FileExistsError, match="already present at"):
        pdjl.immortalize_job(dj)


def test_persistent_download_joblist_skip_queued_resources():
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_dj_list_"))
    pdjl_path = td / "joblistdir"
    ds_dict = common.make_dataset_for_download()
    pdjl = PersistentDownloadJobList(pdjl_path)
    joblist1 = DownloadQueue(api=api)
    resource_id = ds_dict["resources"][0]["id"]
    dj = joblist1.new_job(resource_id=resource_id,
                          download_path=td)
    pdjl.immortalize_job(dj)

    dq = DownloadQueue(api=api, path_persistent_job_list=pdjl_path)
    dq.daemon_download.join()
    dq.daemon_verify.join()

    assert len(dq) == 1
    assert dq.jobs_eternal.num_queued == 1

    # sanity check
    assert pdjl.is_job_queued(resource_id)

    same_job = DownloadJob.from_download_job_state(dj.__getstate__(), api=api)
    dq.add_job(same_job)
    assert len(dq) == 1
    assert dq.jobs_eternal.num_queued == 1



import pathlib

import pytest

from dcoraid.api import (APIConflictError, APINotFoundError, dataset_create,
                         dataset_draft_remove, resource_add, resource_exists)

import common


dpath = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"


def test_dataset_create_same_resource():
    """There should be an error when a resource is added twice"""
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="create-with-same-resource")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict,
                          api=api)
    resource_add(dataset_id=data["id"],
                 path=dpath,
                 api=api
                 )
    with pytest.raises(APIConflictError):
        # Should not be able to upload same resource twice
        resource_add(dataset_id=data["id"],
                     path=dpath,
                     exist_ok=False,
                     api=api
                     )


def test_dataset_create_same_resource_exist_ok():
    """There should be an error when a resource is added twice"""
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="create-with-same-resource")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict,
                          api=api)
    resource_add(dataset_id=data["id"],
                 path=dpath,
                 api=api
                 )

    # Should not be able to upload same resource twice
    resource_add(dataset_id=data["id"],
                 path=dpath,
                 exist_ok=True,
                 api=api,
                 resource_dict={"sp:chip:channel width": 21.0}
                 )

    pkg_dict = api.get("package_show", id=data["id"])
    res_names = [r["name"] for r in pkg_dict["resources"]]
    idres = res_names.index(dpath.name)
    assert len(pkg_dict["resources"]) == 1
    assert pkg_dict["resources"][idres]["sp:chip:channel width"] == 21.0


def test_dataset_creation():
    """Just test whether we can create (and remove) a draft dataset"""
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="basic_test")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict,
                          api=api,
                          )
    # simple test
    assert "authors" in data
    assert data["authors"] == common.USER_NAME
    assert data["state"] == "draft"
    # remove draft dataset
    dataset_draft_remove(dataset_id=data["id"],
                         api=api,
                         )
    with pytest.raises(APINotFoundError):
        # make sure it is gone
        api.get("package_show", id=data["id"])


def test_dataset_creation_wrong_resource_supplement():
    """Pass an invalid resource supplement and see if it fails"""
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="basic_test")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict,
                          api=api,
                          )
    # simple test
    with pytest.raises(APIConflictError):
        resource_add(dataset_id=data["id"],
                     resource_dict={
                             "sp:chip:production date": "2020-15-31"},
                     path=dpath,
                     api=api
                     )


def test_dataset_resource_exists():
    """There should be an error when a resource is added twice"""
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="create-with-same-resource")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict,
                          api=api)
    resource_add(dataset_id=data["id"],
                 path=dpath,
                 api=api
                 )

    assert resource_exists(dataset_id=data["id"],
                           resource_name=dpath.name,
                           api=api)

    assert not resource_exists(dataset_id=data["id"],
                               resource_name=dpath.name,
                               resource_dict={
                                       "sp:chip:channel width": 21.0
                                   },
                               api=api)


def test_dataset_resource_exists2():
    """There should be an error when a resource is added twice"""
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="create-with-same-resource")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict,
                          api=api)
    resource_add(dataset_id=data["id"],
                 path=dpath,
                 resource_dict={"sp:chip:channel width": 21.0},
                 api=api,
                 )

    assert resource_exists(dataset_id=data["id"],
                           resource_name=dpath.name,
                           resource_dict={
                                   "sp:chip:channel width": 21.0
                               },
                           api=api)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import functools
import json
import os
import pathlib
import shutil
import tempfile
import time

from dcoraid.api import CKANAPI, dataset_create
from dcoraid.upload import UploadQueue


CIRCLE = "dcoraid-circle"
COLLECTION = "dcoraid-collection"
DATASET = "dcoraid-dataset"
SERVER = "dcor-dev.mpl.mpg.de"
USER = "dcoraid"
USER_NAME = "DCOR-Aid tester"
TITLE = "DCOR-Aid test dataset"

dpath = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"


def get_api():
    api = CKANAPI(server=SERVER, api_key=get_api_key(), ssl_verify=True)
    return api


def get_api_key():
    key = os.environ.get("DCOR_API_KEY")
    if not key:
        # local file
        kp = pathlib.Path(__file__).parent / "api_key"
        if not kp.exists():
            raise ValueError("No DCOR_API_KEY variable or api_key file!")
        key = kp.read_text().strip()
    return key


def make_dataset_dict(hint=""):
    hint += " "
    dataset_dict = {
        "title": "A test dataset {}{}".format(hint, time.time()),
        "private": True,
        "license_id": "CC0-1.0",
        "owner_org": CIRCLE,
        "authors": USER_NAME,
    }
    return dataset_dict


@functools.lru_cache()
def make_dataset_for_download(seed=0):
    """Set `seed` to get a new dataset (in case you need fresh resource ids)"""
    api = get_api()
    # create some metadata
    dataset_dict = make_dataset_dict(hint="test-download-dataset")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict, api=api)
    joblist = UploadQueue(api=api)
    joblist.new_job(dataset_id=data["id"],
                    paths=[dpath])
    wait_for_job(joblist, data["id"])
    return api.get("package_show", id=data["id"])


def make_upload_task(task_id="123456789",
                     dataset_id=None,
                     dataset_dict=True,
                     resource_paths=[dpath],
                     resource_names=["gorgonzola.rtdc"],
                     resource_supplements=None,
                     ):
    """Return path to example task file"""
    if dataset_dict and not isinstance(dataset_dict, dict):
        dataset_dict = make_dataset_dict(hint="task_test")
    if dataset_dict and dataset_id is None:
        dataset_id = dataset_dict.get("id")
    td = pathlib.Path(tempfile.mkdtemp(prefix="task_"))
    # copy resources there
    new_resource_paths = []
    for pp in resource_paths:
        newpp = td / pathlib.Path(pp).name
        if pathlib.Path(pp).exists():
            # only move the file if the test uses an actual file
            shutil.copy2(pp, newpp)
            new_resource_paths.append(newpp)
        else:
            new_resource_paths.append(pp)
    uj_state = {
        "dataset_id": dataset_id,
        "task_id": task_id,
        "resource_paths": [str(pp) for pp in new_resource_paths],
        "resource_names": resource_names,
        "resource_supplements": resource_supplements,
    }
    data = {"upload_job": uj_state}
    if dataset_dict:
        data["dataset_dict"] = dataset_dict
    taskp = td / "test.dcoraid-task"
    with taskp.open("w") as fd:
        json.dump(data, fd,
                  ensure_ascii=False,
                  indent=2,
                  sort_keys=True)
    return str(taskp)


def wait_for_job(upload_queue, dataset_id, wait_time=60):
    uq = upload_queue
    uj = uq.get_job(dataset_id)
    # wait for the upload to finish
    for _ in range(wait_time*10):
        if uj.state == "done":
            if uq.jobs_eternal:
                # TODO:
                # We do this manually here. Actually, a better solution would
                # be to implement a signal-slot type of workflow where the
                # job tells the queue when it is done.
                uq.jobs_eternal.set_job_done(dataset_id)
            break
        time.sleep(.1)
    else:
        assert False, f"Job '{uj}' not done in {wait_time}s, state {uj.state}!"


def wait_for_job_no_queue(upload_job, wait_time=120):
    uj = upload_job
    # wait for the upload to finish
    for _ in range(wait_time*10):
        uj.task_verify_resources()
        if uj.state == "done":
            break
        time.sleep(.1)
    else:
        assert False, f"Job '{uj}' not done in {wait_time}s, state {uj.state}!"



import pathlib
import tempfile
import uuid

import pytest

from dcoraid.api import dataset_create
from dcoraid.upload import UploadQueue
from dcoraid.upload.queue import DCORAidQueueWarning, PersistentUploadJobList
from dcoraid.upload.task import load_task

import common


data_path = pathlib.Path(__file__).parent / "data"
dpath = data_path / "calibration_beads_47.rtdc"


def test_queue_basic_functionalities():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="create-with-resource")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict, api=api)
    joblist = UploadQueue(api=api)
    uj = joblist.new_job(dataset_id=data["id"],
                         paths=[dpath])
    assert uj.dataset_id == data["id"]
    assert uj in joblist

    samejob = joblist.get_job(data["id"])
    assert samejob is uj

    with pytest.raises(KeyError, match="Job 'peter' not found!"):
        joblist.get_job("peter")

    samejob2 = joblist[0]
    assert samejob2 is uj


def test_queue_create_dataset_with_resource():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="create-with-resource")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict, api=api)
    joblist = UploadQueue(api=api)
    joblist.new_job(dataset_id=data["id"],
                    paths=[dpath])
    common.wait_for_job(joblist, data["id"])


def test_queue_find_zombie_caches():
    api = common.get_api()
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="create-with-resource")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict, api=api)
    # post dataset creation request
    cache_dir = tempfile.mkdtemp("dcoraid_test_upload_cache_")
    fakecache = pathlib.Path(cache_dir) / f"compress-{uuid.uuid4()}"
    fakecache.mkdir()
    realcache = pathlib.Path(cache_dir) / f"compress-{data['id']}"
    joblist = UploadQueue(api=api, cache_dir=cache_dir)
    # disable all daemons, so no uploading happens
    joblist.daemon_compress.join()
    joblist.daemon_upload.join()
    joblist.daemon_verify.join()
    uj = joblist.new_job(
        dataset_id=data["id"],
        paths=[data_path / "calibration_beads_47_nocomp.rtdc"])
    assert uj.state == "init"
    zombies = joblist.find_zombie_caches()
    assert fakecache in zombies
    assert realcache not in zombies
    assert len(zombies) == 1


def test_queue_remove_job():
    """Remove a job from the queue and from the persistent list"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    # create some metadata
    dataset_dict = common.make_dataset_dict(hint="create-with-resource")
    # post dataset creation request
    data = dataset_create(dataset_dict=dataset_dict, api=api)
    joblist = UploadQueue(api=api,
                          path_persistent_job_list=pujl_path)
    # disable all daemons, so no uploading happens
    joblist.daemon_compress.join()
    joblist.daemon_upload.join()
    joblist.daemon_verify.join()
    uj = joblist.new_job(dataset_id=data["id"],
                         paths=[dpath])
    assert uj.state == "init"
    joblist.remove_job(uj.dataset_id)
    assert uj not in joblist
    assert not joblist.jobs_eternal.job_exists(uj.dataset_id)

    # adding it again should work
    uj2 = joblist.new_job(dataset_id=data["id"],
                          paths=[dpath])
    assert uj2 in joblist
    assert uj2.__getstate__() == uj.__getstate__()


def test_persistent_upload_joblist_basic():
    """basic job tests"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = common.make_upload_task()
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path, api=api)

    # add a job
    pujl.immortalize_job(uj)
    assert uj in pujl
    assert uj.dataset_id in pujl

    # find that job
    uj_same = pujl.summon_job(uj.dataset_id, api=api)
    assert uj_same is not uj, "not same instance"
    assert uj_same.__getstate__() == uj.__getstate__(), "same data"
    ids = pujl.get_queued_dataset_ids()
    assert uj.dataset_id in ids

    # remove a job
    assert pujl.job_exists(uj.dataset_id)
    assert pujl.is_job_queued(uj.dataset_id)
    assert not pujl.is_job_done(uj.dataset_id)
    pujl.obliterate_job(uj.dataset_id)
    assert uj not in pujl
    assert not pujl.job_exists(uj.dataset_id)


def test_persistent_upload_joblist_job_added_in_queue():
    """Test whether queuing works"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = common.make_upload_task()
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path, api=api)

    uq = UploadQueue(api=api, path_persistent_job_list=pujl_path)
    uq.daemon_compress.join()
    uq.daemon_upload.join()
    uq.daemon_verify.join()

    assert pujl.num_queued == 0
    uq.add_job(uj)
    assert pujl.num_queued == 1
    assert pujl.num_completed == 0


def test_persistent_upload_joblist_job_added_and_finished_in_queue():
    """Test whether queuing works"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = common.make_upload_task()
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path, api=api)

    uq = UploadQueue(api=api, path_persistent_job_list=pujl_path)

    assert pujl.num_queued == 0
    uq.add_job(uj)

    common.wait_for_job(uq, uj.dataset_id)

    assert pujl.num_queued == 0
    assert pujl.num_completed == 1


def test_persistent_upload_joblist_done():
    """test things when a job is done"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = common.make_upload_task()
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path, api=api)
    pujl.immortalize_job(uj)
    pujl.set_job_done(uj.dataset_id)
    assert pujl.job_exists(uj.dataset_id)
    assert not pujl.is_job_queued(uj.dataset_id)
    assert pujl.is_job_done(uj.dataset_id)
    assert uj in pujl

    ids = pujl.get_queued_dataset_ids()
    assert uj.dataset_id not in ids


def test_persistent_upload_joblist_error_exists():
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = common.make_upload_task()
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path, api=api)
    pujl.immortalize_job(uj)
    with pytest.raises(FileExistsError, match="already present at"):
        pujl.immortalize_job(uj)


def test_persistent_upload_joblist_skip_finished_resources():
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = pathlib.Path(common.make_upload_task())
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path,
                   api=api,
                   update_dataset_id=True)
    pujl.immortalize_job(uj)

    uq = UploadQueue(api=api, path_persistent_job_list=pujl_path)
    assert len(uq) == 1
    assert uq.jobs_eternal.num_queued == 1

    common.wait_for_job(uq, uj.dataset_id)

    # sanity check
    assert pujl.is_job_done(uj.dataset_id)

    # try to add the task again
    same_job = load_task(task_path,
                         api=api,
                         update_dataset_id=True)
    uq.add_job(same_job)
    assert len(uq) == 1
    assert uq.jobs_eternal.num_queued == 0


def test_persistent_upload_joblist_skip_missing_resources():
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = pathlib.Path(common.make_upload_task())
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path,
                   api=api,
                   update_dataset_id=True)
    pujl.immortalize_job(uj)

    # hide the original data directory so loading the task fails
    task_dir = task_path.parent
    temp_task_dir = task_path.parent.with_name("test_rename_temp")
    task_dir.rename(temp_task_dir)
    assert not task_dir.exists()
    with pytest.warns(DCORAidQueueWarning, match="resources are missing"):
        uq = UploadQueue(api=api, path_persistent_job_list=pujl_path)
    uq.daemon_compress.join()
    uq.daemon_upload.join()
    uq.daemon_verify.join()
    # sanity checks
    assert len(uq) == 0
    assert uq.jobs_eternal.num_queued == 1
    # now add the new task
    same_job = load_task(temp_task_dir / task_path.name,
                         api=api,
                         update_dataset_id=True)
    uq.add_job(same_job)
    assert len(uq) == 1
    assert uq.jobs_eternal.num_queued == 1


def test_persistent_upload_joblist_skip_queued_resources():
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = pathlib.Path(common.make_upload_task())
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path,
                   api=api,
                   update_dataset_id=True)
    pujl.immortalize_job(uj)

    uq = UploadQueue(api=api, path_persistent_job_list=pujl_path)
    uq.daemon_compress.join()
    uq.daemon_upload.join()
    uq.daemon_verify.join()

    assert len(uq) == 1
    assert uq.jobs_eternal.num_queued == 1

    # sanity check
    assert pujl.is_job_queued(uj.dataset_id)

    # try to add the task again
    same_job = load_task(task_path,
                         api=api,
                         update_dataset_id=True)
    uq.add_job(same_job)
    assert len(uq) == 1
    assert uq.jobs_eternal.num_queued == 1


def test_persistent_upload_joblist_error_exists_done():
    """test things when a job is done"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = common.make_upload_task()
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path, api=api)
    pujl.immortalize_job(uj)
    pujl.set_job_done(uj.dataset_id)
    with pytest.raises(FileExistsError, match="is already done"):
        pujl.immortalize_job(uj)


def test_persistent_upload_joblist_warning_dataset_deleted_on_server():
    """warning when dataset has been deleted on server"""
    api = common.get_api()
    td = pathlib.Path(tempfile.mkdtemp(prefix="persistent_uj_list_"))
    pujl_path = td / "joblistdir"
    task_path = common.make_upload_task()
    pujl = PersistentUploadJobList(pujl_path)
    uj = load_task(task_path, api=api)
    pujl.immortalize_job(uj)

    # Now change the dataset ID of the job
    new_id = str(uuid.uuid4())
    pj_path_old = pujl.path_queued / (uj.dataset_id + ".json")
    pj_path_new = pujl.path_queued / (new_id + ".json")

    json_data = pj_path_old.read_text(encoding="utf-8")
    json_data = json_data.replace(uj.dataset_id, new_id)
    pj_path_new.write_text(json_data, encoding="utf-8")
    pj_path_old.unlink()

    with pytest.warns(UserWarning, match=f"{new_id} could not be found"):
        UploadQueue(api=api, path_persistent_job_list=pujl_path)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from functools import lru_cache
import os.path as os_path
import pathlib
import pkg_resources

from PyQt5 import uic, QtCore, QtWidgets
from PyQt5.QtCore import QStandardPaths

from ...common import ConnectionTimeoutErrors
from ...upload import UploadQueue, task

from ..api import get_ckan_api
from ..tools import ShowWaitCursor

from . import circle_mgr
from .dlg_upload import NoCircleSelectedError, UploadDialog
from .widget_tablecell_actions import TableCellActions


class UploadWidget(QtWidgets.QWidget):
    upload_finished = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        """Manage running uploads
        """
        super(UploadWidget, self).__init__(*args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "dcoraid.gui.upload", "widget_upload.ui")
        uic.loadUi(path_ui, self)

        self._dlg_manual = None

        # button for adding new dataset manually
        self.toolButton_new_upload.clicked.connect(self.on_upload_manual)

        # menu button for adding tasks
        menu = QtWidgets.QMenu()
        act1 = QtWidgets.QAction("Select task files from disk", self)
        act1.setData("single")
        menu.addAction(act1)
        act2 = QtWidgets.QAction(
            "Recursively find and load task files from a folder", self)
        act2.setData("bulk")
        menu.addAction(act2)
        menu.triggered.connect(self.on_upload_task)
        self.toolButton_load_upload_tasks.setMenu(menu)

        # Underlying upload class
        # use a persistent shelf to be able to resume uploads on startup
        shelf_path = os_path.join(
            QStandardPaths.writableLocation(
                QStandardPaths.AppLocalDataLocation),
            "persistent_upload_jobs")
        self.cache_dir = QtCore.QStandardPaths.writableLocation(
            QtCore.QStandardPaths.CacheLocation)

        try:
            self.jobs = UploadQueue(api=get_ckan_api(),
                                    path_persistent_job_list=shelf_path,
                                    cache_dir=self.cache_dir)
        except ConnectionTimeoutErrors:
            # TODO: allow user to re-enable without restarting DCOR-Aid
            self.jobs = None
            self.setEnabled(False)
        else:
            self.widget_jobs.set_job_list(self.jobs)
            # upload finished signal
            self.widget_jobs.upload_finished.connect(self.upload_finished)

    @QtCore.pyqtSlot()
    def on_upload_manual(self):
        """Guides the user through the process of creating a dataset

        A draft dataset is created to which the resources are then
        uploaded.
        """
        try:
            self._dlg_manual = UploadDialog(self)
        except NoCircleSelectedError:
            self._dlg_manual = None
        else:
            self._dlg_manual.finished.connect(self.on_upload_manual_ready)
            self._dlg_manual.close()
            self._dlg_manual.exec()

    @QtCore.pyqtSlot(object)
    def on_upload_manual_ready(self, upload_dialog):
        """Proceed with resource upload as defined by `update_dialog`

        `update_dialog` is an instance of `dlg_upload.UploadDialog`
        and contains all information necessary to run the resource
        upload. Supplementary resource metadata is extracted here
        as well.
        """
        # Remove all magic keys from the schema data (they are
        # only used internally by DCOR-Aid and don't belong on DCOR).
        rdata = upload_dialog.rvmodel.get_all_data(magic_keys=False)
        paths = []
        names = []
        supps = []
        for path in rdata:
            paths.append(pathlib.Path(path))
            names.append(rdata[path]["file"]["filename"])
            supps.append(rdata[path]["supplement"])
        dataset_dict = upload_dialog.dataset_dict
        # add the entry to the job list
        self.jobs.new_job(dataset_id=dataset_dict["id"],
                          paths=paths,
                          resource_names=names,
                          supplements=supps)

    @QtCore.pyqtSlot(QtWidgets.QAction)
    def on_upload_task(self, action):
        """Import an UploadJob task file and add it to the queue

        This functionality is mainly used for automation. Another
        software creates upload tasks which are then loaded by
        DCOR-Aid.
        """
        if action.data() == "single":
            files, _ = QtWidgets.QFileDialog.getOpenFileNames(
                self,
                "Select DCOR-Aid task files",
                ".",
                "DCOR-Aid task files (*.dcoraid-task)",
            )
        else:
            tdir = QtWidgets.QFileDialog.getExistingDirectory(
                self,
                "Select folder to search for DCOR-Aid task files",
                ".",
                QtWidgets.QFileDialog.ShowDirsOnly,
            )
            files = pathlib.Path(tdir).rglob("*.dcoraid-task")

        # Keep track of a persistent task ID to dataset ID dictionary
        path_id_dict = os_path.join(
            QStandardPaths.writableLocation(
                QStandardPaths.AppLocalDataLocation),
            "map_task_to_dataset_id.txt")
        map_task_to_dataset_id = task.PersistentTaskDatasetIDDict(path_id_dict)
        api = get_ckan_api()
        settings = QtCore.QSettings()
        update_dataset_id = bool(int(settings.value(
            "uploads/update task with dataset id", "1")))
        dataset_kwargs = {}
        jobs_known_count = 0
        jobs_imported_count = 0
        jobs_total_count = 0
        for pp in files:
            jobs_total_count += 1
            if (not task.task_has_circle(pp)
                    and "owner_org" not in dataset_kwargs):
                # Let the user choose a circle.
                # Note the above test for "owner_org": The user only
                # chooses the circle *once* for *all* task files.
                cdict = circle_mgr.request_circle(self)
                if cdict is None:
                    # The user aborted, so we won't continue!
                    break
                else:
                    dataset_kwargs["owner_org"] = cdict["name"]
            with ShowWaitCursor():
                upload_job = task.load_task(
                    path=pp,
                    map_task_to_dataset_id=map_task_to_dataset_id,
                    api=api,
                    dataset_kwargs=dataset_kwargs,
                    update_dataset_id=update_dataset_id,
                    cache_dir=self.cache_dir,
                )
                job_msg = self.jobs.add_job(upload_job)
                if job_msg == "new":
                    jobs_imported_count += 1
                else:
                    jobs_known_count += 1

        # sanity check
        assert jobs_known_count + jobs_imported_count == jobs_total_count
        if jobs_known_count:
            # Display a message box telling the user that this job is known
            QtWidgets.QMessageBox.information(
                self,
                "Task import complete",
                f"{jobs_known_count} out of {jobs_total_count} tasks were "
                + "already imported in a previous DCOR-Aid session."
            )


class UploadTableWidget(QtWidgets.QTableWidget):
    upload_finished = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super(UploadTableWidget, self).__init__(*args, **kwargs)
        self.jobs = []  # Will become UploadQueue with self.set_job_list

        settings = QtCore.QSettings()
        if bool(int(settings.value("debug/without timers", "0"))):
            self.timer = None
        else:
            self.timer = QtCore.QTimer()
            self.timer.timeout.connect(self.update_job_status)
            self.timer.start(1000)
        self._finished_uploads = []

    def set_job_list(self, jobs):
        """Set the current job list

        The job list can be a `list`, but it is actually
        an `UploadQueue`.
        """
        # This is the actual initialization
        self.jobs = jobs

    @QtCore.pyqtSlot(str)
    def on_job_abort(self, dataset_id):
        self.jobs.abort_job(dataset_id)

    @QtCore.pyqtSlot(str)
    def on_job_delete(self, dataset_id):
        self.jobs.remove_job(dataset_id)

    @QtCore.pyqtSlot(str)
    def on_upload_finished(self, dataset_id):
        """Triggers upload_finished whenever an upload is finished"""
        if dataset_id not in self._finished_uploads:
            self._finished_uploads.append(dataset_id)
            self.jobs.jobs_eternal.set_job_done(dataset_id)
            self.upload_finished.emit()

    @QtCore.pyqtSlot()
    def update_job_status(self):
        """Update UI with information from self.jobs (UploadJobList)"""
        # disable updates
        self.setUpdatesEnabled(False)
        # make sure the length of the table is long enough
        self.setRowCount(len(self.jobs))
        self.setColumnCount(6)

        for row, job in enumerate(self.jobs):
            status = job.get_status()
            self.set_label_item(row, 0, job.dataset_id[:5])
            try:
                title = get_dataset_title(job.dataset_id)
            except BaseException:
                # Probably a connection error
                title = "-- error getting dataset title --"
            self.set_label_item(row, 1, title)
            self.set_label_item(row, 2, status["state"])
            self.set_label_item(row, 3, job.get_progress_string())
            self.set_label_item(row, 4, job.get_rate_string())
            if status["state"] == "done":
                self.on_upload_finished(job.dataset_id)
            self.set_actions_item(row, 5, job)

        # spacing (did not work in __init__)
        header = self.horizontalHeader()
        header.setSectionResizeMode(
            0, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.Stretch)
        # enable updates again
        self.setUpdatesEnabled(True)

    def set_label_item(self, row, col, label):
        """Get/Create a Qlabel at the specified position

        User has to make sure that row and column count are set
        """
        label = "{}".format(label)
        item = self.item(row, col)
        if item is None:
            item = QtWidgets.QTableWidgetItem(label)
            item.setFlags(QtCore.Qt.ItemIsEnabled)
            self.setItem(row, col, item)
        else:
            if item.text() != label:
                item.setText(label)

    def set_actions_item(self, row, col, job):
        """Set/Create a TableCellActions widget in the table

        Refreshes the widget and also connects signals.
        """
        wid = self.cellWidget(row, col)
        if wid is None:
            wid = TableCellActions(job)
            wid.delete_job.connect(self.on_job_delete)
            wid.abort_job.connect(self.on_job_abort)
            self.setCellWidget(row, col, wid)
        wid.refresh_visibility(job)


@lru_cache(maxsize=10000)
def get_dataset_title(dataset_id):
    api = get_ckan_api()
    ddict = api.get("package_show", id=dataset_id)
    return ddict["title"]



import pathlib
import random

import pytest

from dcoraid.api import dataset_create
from dcoraid.dbmodel import db_api

import common

dpath = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"


def test_get_circles():
    api = common.get_api()
    db = db_api.APIInterrogator(api=api)
    circles = db.get_circles()
    assert common.CIRCLE in circles
    # requires that the "dcoraid" user is in the figshare-import circle
    assert "figshare-import" in circles


def test_get_collections():
    api = common.get_api()
    db = db_api.APIInterrogator(api=api)
    collections = db.get_collections()
    assert common.COLLECTION in collections
    # requires that the "dcoraid" user is in the figshare-collection collection
    assert "figshare-collection" in collections


def test_public_api_interrogator():
    """This test uses the figshare datasets on SERVER"""
    api = common.get_api()
    db = db_api.APIInterrogator(api=api)
    assert common.CIRCLE in db.get_circles()
    assert common.COLLECTION in db.get_collections()
    assert common.USER in db.get_users()


def test_user_data():
    """Test the user information"""
    api = common.get_api()
    db = db_api.APIInterrogator(api=api)
    data = db.user_data
    assert data["fullname"] == common.USER_NAME, "fullname not correct"


def test_search_dataset_basic():
    api = common.get_api()
    ranstr = ''.join(random.choice("0123456789") for _i in range(10))
    # Create a test dataset
    dataset_create({"title": "{} {}".format(common.TITLE, ranstr),
                    "owner_org": common.CIRCLE,
                    "authors": common.USER_NAME,
                    "license_id": "CC0-1.0",
                    "groups": [{"name": common.COLLECTION}],
                    },
                   api=api,
                   resources=[dpath],
                   activate=True)
    db = db_api.APIInterrogator(api=api)
    # Positive test
    data = db.search_dataset(query="dcoraid",
                             circles=[common.CIRCLE],
                             collections=[common.COLLECTION],
                             )
    assert len(data) >= 1
    for dd in data:
        if dd["name"].endswith(ranstr):
            break
    else:
        assert False, "{} not found!".format(common.DATASET)
    # Negative test
    data = db.search_dataset(query="cliauwenlc_should_never_exist",
                             circles=[common.CIRCLE],
                             collections=[common.COLLECTION],
                             )
    assert len(data) == 0, "search result for non-existent dataset?"


def test_search_dataset_limit():
    api = common.get_api()
    ranstr = ''.join(random.choice("0123456789") for _i in range(10))
    dataset_ids = []
    # Create three test datasets
    for _ in range(3):
        ds_dict = dataset_create(
            {"title": "{} {}".format(common.TITLE, ranstr),
             "owner_org": common.CIRCLE,
             "authors": common.USER_NAME,
             "license_id": "CC0-1.0",
             "groups": [{"name": common.COLLECTION}],
             },
            api=api,
            resources=[dpath],
            activate=True)
        dataset_ids.append(ds_dict["id"])
    db = db_api.APIInterrogator(api=api)
    data_limited = db.search_dataset(query=ranstr,
                                     circles=[common.CIRCLE],
                                     collections=[common.COLLECTION],
                                     circle_collection_union=True,
                                     limit=2
                                     )
    assert len(data_limited) == 2
    data_unlimited = db.search_dataset(query=ranstr,
                                       circles=[common.CIRCLE],
                                       collections=[common.COLLECTION],
                                       circle_collection_union=True,
                                       limit=0
                                       )
    assert len(data_unlimited) == 3


def test_search_dataset_limit_negative_error():
    api = common.get_api()
    ranstr = ''.join(random.choice("0123456789") for _i in range(10))
    # Create three test datasets
    dataset_create(
        {"title": "{} {}".format(common.TITLE, ranstr),
         "owner_org": common.CIRCLE,
         "authors": common.USER_NAME,
         "license_id": "CC0-1.0",
         "groups": [{"name": common.COLLECTION}],
         },
        api=api,
        resources=[dpath],
        activate=True)
    db = db_api.APIInterrogator(api=api)
    with pytest.raises(ValueError, match="must be 0 or >0"):
        db.search_dataset(query=ranstr,
                          circles=[common.CIRCLE],
                          collections=[common.COLLECTION],
                          circle_collection_union=True,
                          limit=-1
                          )


def test_search_dataset_only_one_filter_query():
    api = common.get_api()
    db = db_api.APIInterrogator(api=api)
    ds = db.search_dataset(filter_queries=[f"-creator_user_id:{api.user_id}"])
    for di in ds:
        if di["name"] == "figshare-7771184-v2":
            break
    else:
        assert False, "Search did not return figshare-7771184-v2!"


def test_get_datasets_user_shared_figshare():
    """The figshare circle must have the user "dcoraid" as a member
    """
    api = common.get_api()
    db = db_api.APIInterrogator(api=api)
    datasets = db.get_datasets_user_shared()
    for dd in datasets:
        if dd["id"] == "89bf2177-ffeb-9893-83cc-b619fc2f6663":
            break
    else:
        assert False, "Search did not return figshare-7771184-v2!"
    assert len(datasets) == 1


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "_call_"):
            loc[key]()



import pathlib
import tempfile
import time
from unittest import mock

import pytest
import dclab.cli
from dcoraid.api import dataset_create
from dcoraid.upload import job

import common


data_path = pathlib.Path(__file__).parent / "data"
rtdc_paths = [data_path / "calibration_beads_47.rtdc",
              data_path / "calibration_beads_47_nocomp.rtdc",
              ]


def test_initialize():
    api = common.get_api()
    # create some metadata
    bare_dict = common.make_dataset_dict(hint="create-with-resource")
    # create dataset (to get the "id")
    dataset_dict = dataset_create(dataset_dict=bare_dict, api=api)
    uj = job.UploadJob(api=api, dataset_id=dataset_dict["id"],
                       resource_paths=rtdc_paths)
    assert uj.state == "init"


def test_full_upload():
    api = common.get_api()
    # create some metadata
    bare_dict = common.make_dataset_dict(hint="create-with-resource")
    # create dataset (to get the "id")
    dataset_dict = dataset_create(dataset_dict=bare_dict, api=api)
    uj = job.UploadJob(api=api, dataset_id=dataset_dict["id"],
                       resource_paths=rtdc_paths)
    assert uj.state == "init"
    uj.task_compress_resources()
    assert uj.state == "parcel"
    uj.task_upload_resources()
    assert uj.state == "online"

    common.wait_for_job_no_queue(uj)


def test_saveload():
    api = common.get_api()
    # create some metadata
    bare_dict = common.make_dataset_dict(hint="create-with-resource")
    # create dataset (to get the "id")
    dataset_dict = dataset_create(dataset_dict=bare_dict, api=api)
    uj = job.UploadJob(api=api, dataset_id=dataset_dict["id"],
                       resource_paths=rtdc_paths, task_id="hanspeter")
    state = uj.__getstate__()
    assert state["dataset_id"] == dataset_dict["id"]
    assert rtdc_paths[0].samefile(state["resource_paths"][0])
    assert rtdc_paths[0].name == state["resource_names"][0]

    # now create a new job from the state
    uj2 = job.UploadJob.from_upload_job_state(state, api=api)
    state2 = uj2.__getstate__()
    assert state2["dataset_id"] == dataset_dict["id"]
    assert rtdc_paths[0].samefile(state2["resource_paths"][0])
    assert rtdc_paths[0].name == state2["resource_names"][0]
    assert state2["task_id"] == "hanspeter"


@mock.patch.object(job.shutil, "disk_usage")
def test_state_compress_disk_wait(disk_usage_mock):
    """If not space left on disk, upload job goes to state "disk-wait"""""
    api = common.get_api()
    # create some metadata
    bare_dict = common.make_dataset_dict(hint="create-with-resource")
    # create dataset (to get the "id")
    dataset_dict = dataset_create(dataset_dict=bare_dict, api=api)
    uj = job.UploadJob(api=api, dataset_id=dataset_dict["id"],
                       resource_paths=[rtdc_paths[1]])
    assert uj.state == "init"

    class DiskUsageReturn1:
        free = 0

    class DiskUsageReturn2:
        free = 1024**3

    disk_usage_mock.side_effect = [DiskUsageReturn1, DiskUsageReturn2]

    t0 = time.perf_counter()
    uj.task_compress_resources()
    t1 = time.perf_counter()
    assert t1-t0 > 0.15, "function waits 0.2s in-between checks"
    disk_usage_mock.assert_called()


@mock.patch.object(job.shutil, "disk_usage")
def test_state_compress_remove_other_upload_jobs(disk_usage_mock):
    """
    A job may delete rogue upload cache data from jobs that would
    upload after it.
    """
    api = common.get_api()
    cache_dir = pathlib.Path(tempfile.mkdtemp(prefix="upload_job_cache_test_"))
    # create some metadata
    test_list = []
    for _ in range(3):
        bare_dict = common.make_dataset_dict(hint="create-with-resource")
        dataset_dict = dataset_create(dataset_dict=bare_dict, api=api)
        uj = job.UploadJob(api=api,
                           dataset_id=dataset_dict["id"],
                           resource_paths=[rtdc_paths[1]],
                           cache_dir=cache_dir,
                           )
        test_list.append((dataset_dict["id"], uj, dataset_dict))

    # sort according to dataset ID
    test_list = sorted(test_list)
    uj1, uj2, uj3 = [td[1] for td in test_list]

    class DiskUsageReturn1:
        free = 0

    class DiskUsageReturn2:
        free = 1024**3

    disk_usage_mock.side_effect = [
        DiskUsageReturn2,  # uj1
        DiskUsageReturn2,  # uj3
        DiskUsageReturn1,  # uj2 fail
        DiskUsageReturn2]  # uj2 success

    # compress the first and the third dataset
    uj1.task_compress_resources()
    uj3.task_compress_resources()
    assert uj1.cache_dir.exists()
    assert not uj2.cache_dir.exists()
    assert uj3.cache_dir.exists()

    # Now compress uj2 and tell it that the disk is full. It should delete
    # the cache_dir of uj3, but not of uj1.
    uj2.task_compress_resources()
    assert uj1.cache_dir.exists()
    assert uj2.cache_dir.exists()
    assert not uj3.cache_dir.exists()


def test_state_compress_reuse():
    """Reuse previously compressed datasets"""""
    api = common.get_api()
    # create some metadata
    bare_dict = common.make_dataset_dict(hint="create-with-resource")
    # create dataset (to get the "id")
    dataset_dict = dataset_create(dataset_dict=bare_dict, api=api)
    uj = job.UploadJob(api=api, dataset_id=dataset_dict["id"],
                       resource_paths=[rtdc_paths[1]])
    assert uj.state == "init"

    with mock.patch.object(job, "compress",
                           side_effect=dclab.cli.compress) as mockobj:
        uj.task_compress_resources()
        mockobj.assert_called_once()

    with mock.patch.object(job, "compress",
                           side_effect=dclab.cli.compress) as mockobj:
        uj.task_compress_resources()
        with pytest.raises(AssertionError):
            mockobj.assert_called()


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import tempfile
from unittest import mock

from dcoraid.download import job

import common


def test_initialize():
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    ds_dict = common.make_dataset_for_download()
    dj = job.DownloadJob(api=api,
                         resource_id=ds_dict["resources"][0]["id"],
                         download_path=td)
    assert dj.state == "init"


def test_download_resume():
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    ds_dict = common.make_dataset_for_download()
    dj = job.DownloadJob(api=api,
                         resource_id=ds_dict["resources"][0]["id"],
                         download_path=td)
    dj.task_download_resource()

    # now attempt a resume-download
    # create a truncated temporary file
    orig = dj.path_temp.with_name("orig.rtdc~")
    dj.path_temp.rename(orig)
    with orig.open("rb") as fd_o:
        start_bytes = fd_o.read(100)
    with dj.path_temp.open("wb") as fd_t:
        fd_t.write(start_bytes)

    # try to resume from that temporary file
    dj2 = job.DownloadJob(api=api,
                          resource_id=ds_dict["resources"][0]["id"],
                          download_path=td)
    dj2.task_download_resource()
    dj2.task_verify_resource()
    assert dj2.start_time is not None
    assert dj2.end_time is not None
    assert dj2.path.exists()
    assert dj2.file_bytes_downloaded == dj.file_size - 100
    assert dj2.file_size == dj2.path.stat().st_size


def test_full_download():
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    ds_dict = common.make_dataset_for_download()
    dj = job.DownloadJob(api=api,
                         resource_id=ds_dict["resources"][0]["id"],
                         download_path=td)
    assert dj.state == "init"
    dj.task_download_resource()
    assert dj.state == "downloaded"
    assert dj.path_temp.exists()
    assert not dj.path.exists()
    dj.task_verify_resource()
    assert dj.state == "done"
    assert not dj.path_temp.exists()
    assert dj.path.exists()


def test_full_download_file_exists():
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    ds_dict = common.make_dataset_for_download()
    dj = job.DownloadJob(api=api,
                         resource_id=ds_dict["resources"][0]["id"],
                         download_path=td)
    assert dj.path.is_dir()
    dj.task_download_resource()
    dj.task_verify_resource()
    assert dj.start_time is not None
    assert dj.end_time is not None
    assert dj.path.exists()
    assert not dj.path.is_dir()

    # now attempt to download the file a second time and make
    # sure that it is not actually downloaded
    dj2 = job.DownloadJob(api=api,
                          resource_id=ds_dict["resources"][0]["id"],
                          download_path=td)
    dj2.task_download_resource()
    dj2.task_verify_resource()
    assert dj2.start_time is None
    assert dj2.end_time is None
    assert dj2.path.exists()
    assert dj2.path.samefile(dj.path)


def test_get_status():
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    ds_dict = common.make_dataset_for_download()
    dj = job.DownloadJob(api=api,
                         resource_id=ds_dict["resources"][0]["id"],
                         download_path=td)
    size = dj.get_resource_dict()["size"]
    assert size != 0
    assert dj.get_status()["bytes total"] == size
    assert dj.get_status()["bytes downloaded"] == 0
    assert dj.get_status()["rate"] == 0
    dj.task_download_resource()
    assert dj.get_status()["bytes downloaded"] == size
    assert dj.get_status()["rate"] > 0


def test_saveload():
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    ds_dict = common.make_dataset_for_download()
    dj = job.DownloadJob(api=api,
                         resource_id=ds_dict["resources"][0]["id"],
                         download_path=td)

    state = dj.__getstate__()
    assert state["resource_id"] == ds_dict["resources"][0]["id"]

    # now create a new job from the state
    dj2 = job.DownloadJob.from_download_job_state(state, api=api)
    state2 = dj2.__getstate__()
    assert state2["resource_id"] == ds_dict["resources"][0]["id"]
    assert dj2.path.samefile(state["download_path"])


@mock.patch.object(job.shutil, "disk_usage")
def test_state_init_disk_wait(disk_usage_mock):
    """If not space left on disk, download job goes to state "disk-wait"""""
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    ds_dict = common.make_dataset_for_download()
    dj = job.DownloadJob(api=api,
                         resource_id=ds_dict["resources"][0]["id"],
                         download_path=td)
    assert dj.state == "init"

    class DiskUsageReturn1:
        free = 0

    class DiskUsageReturn2:
        free = 1024**3

    disk_usage_mock.side_effect = [DiskUsageReturn1, DiskUsageReturn2]

    dj.task_download_resource()
    assert dj.state == "wait-disk"
    dj.task_download_resource()
    assert dj.state == "downloaded"
    disk_usage_mock.assert_called()



import pathlib
import warnings

from ..api import APINotFoundError
from ..worker import Daemon

from .job import UploadJob
from .task import LocalTaskResourcesNotFoundError, load_task, save_task


class DCORAidQueueWarning(UserWarning):
    pass


class PersistentUploadJobList:
    def __init__(self, path):
        """A file-system and JSON-based persistent UploadJob list"""
        self.path = pathlib.Path(path)
        self.path_completed = self.path / "completed"
        self.path_queued = self.path / "queued"
        self.path_completed.mkdir(parents=True, exist_ok=True)
        self.path_queued.mkdir(parents=True, exist_ok=True)

    def __contains__(self, item):
        if isinstance(item, UploadJob):
            dataset_id = item.dataset_id
        else:
            dataset_id = item
        return self.job_exists(dataset_id)

    @property
    def num_completed(self):
        """Return number of completed tasks"""
        return len(list(self.path_completed.glob("*.json")))

    @property
    def num_queued(self):
        """Return number of queued tasks"""
        return len(list(self.path_queued.glob("*.json")))

    def get_queued_dataset_ids(self):
        """Return list of DCOR dataset IDs corresponding to queued jobs"""
        return sorted([pp.stem for pp in self.path_queued.glob("*.json")])

    def is_job_done(self, dataset_id):
        jp = self.path_completed / (dataset_id + ".json")
        return jp.exists()

    def is_job_queued(self, dataset_id):
        jp = self.path_queued / (dataset_id + ".json")
        return jp.exists()

    def immortalize_job(self, upload_job):
        """Put this job in the persistent queue list"""
        pout = self.path_queued / (upload_job.dataset_id + ".json")
        if self.is_job_queued(upload_job.dataset_id):
            raise FileExistsError(f"The job '{upload_job.dataset_id}' is "
                                  f"already present at '{pout}'!")
        elif self.is_job_done(upload_job.dataset_id):
            # This is safe in async mode, because checking for "done"
            # is done after checking for queued (above case).
            raise FileExistsError(f"The job '{upload_job.dataset_id}' is "
                                  f"already done!")
        save_task(upload_job=upload_job, path=pout)

    def job_exists(self, dataset_id):
        return self.is_job_queued(dataset_id) or self.is_job_done(dataset_id)

    def obliterate_job(self, dataset_id):
        """Remove a job from the persistent queue list"""
        pdel = self.path_queued / (dataset_id + ".json")
        pdel.unlink()

    def set_job_done(self, dataset_id):
        """Move the job from the queue to the complete list"""
        pin = self.path_queued / (dataset_id + ".json")
        pout = self.path_completed / (dataset_id + ".json")
        pin.rename(pout)

    def summon_job(self, dataset_id, api, cache_dir=None):
        """Instantiate job from the persistent queue list"""
        pin = self.path_queued / (dataset_id + ".json")
        upload_job = load_task(path=pin, api=api, cache_dir=cache_dir)
        assert upload_job.dataset_id == dataset_id
        return upload_job


class UploadQueue:
    def __init__(self, api, path_persistent_job_list=None, cache_dir=None):
        """Manager for running multiple UploadJobs in sequence

        Parameters
        ----------
        api: dcoraid.api.CKANAPI
            The CKAN/DCOR API instance used for the uploads
        path_persistent_job_list: str or pathlib.Path
            Path to a directory for storing UploadJobs in a
            persistent manner across restarts.
        cache_dir: str or pathlib.Path
            Cache directory for storing compressed .rtdc files;
            if not supplied, a temporary directory is created for
            each UploadJob
        """
        self.api = api.copy()
        if not api.api_key:
            warnings.warn("No API key is set! Upload will not work!")
        self.cache_dir = cache_dir
        self.jobs = []
        if path_persistent_job_list is not None:
            self.jobs_eternal = PersistentUploadJobList(
                path_persistent_job_list)
            # add any previously queued jobs
            for dataset_id in self.jobs_eternal.get_queued_dataset_ids():
                try:
                    uj = self.jobs_eternal.summon_job(dataset_id,
                                                      api=self.api,
                                                      cache_dir=self.cache_dir)
                except APINotFoundError:
                    pp = self.jobs_eternal.path_queued / (dataset_id + ".json")
                    warnings.warn(f"Datast {dataset_id} could not be found "
                                  f"on {self.api.server}! If the dataset has "
                                  f"been deleted, please remove the local "
                                  f"file {pp}.",
                                  DCORAidQueueWarning)
                except LocalTaskResourcesNotFoundError as e:
                    resstr = ", ".join([str(pp) for pp in e.missing_resources])
                    warnings.warn("The following resources are missing for "
                                  f"dataset {dataset_id}: {resstr}. The "
                                  "job will not be queued.",
                                  DCORAidQueueWarning)
                else:
                    self.jobs.append(uj)
        else:
            self.jobs_eternal = None
        self.daemon_compress = CompressDaemon(self.jobs)
        self.daemon_upload = UploadDaemon(self.jobs)
        self.daemon_verify = VerifyDaemon(self.jobs)

    def __contains__(self, upload_job):
        return upload_job in self.jobs

    def __getitem__(self, index):
        return self.jobs[index]

    def __len__(self):
        return len(self.jobs)

    def find_zombie_caches(self):
        """Return list of cache directories that don't belong to this Queue

        Returns
        -------
        zombies: list of pathlib.Path
            List of zombie cache directories
        """
        if self.cache_dir is None:
            # We can only check if the directory was given
            raise ValueError("UploadQueue was instantiated without cache_dir!")
        else:
            dataset_ids = [job.dataset_id for job in self]
            zombies = []
            for pp in pathlib.Path(self.cache_dir).glob("compress-*"):
                did = pp.name.split("-", 1)[1]
                if did not in dataset_ids:
                    zombies.append(pp)
        return zombies

    def abort_job(self, dataset_id):
        """Abort a running job but don't remove it from the queue"""
        job = self.get_job(dataset_id)
        if job.state == "transfer":
            job.set_state("abort")
            # https://github.com/requests/toolbelt/issues/297
            self.daemon_upload.terminate()
            self.daemon_upload = UploadDaemon(self.jobs)
        elif job.state == "compress":
            job.set_state("abort")
            self.daemon_compress.terminate()
            self.daemon_compress = CompressDaemon(self.jobs)

    def add_job(self, upload_job):
        """Add an UploadJob to the queue

        Returns
        -------
        job_message: str
            Either "known" (eternal job, not done), "finished"
            (eternal job, done), or "new" (completely new job)
        """
        if self.jobs_eternal is not None:
            if upload_job in self.jobs_eternal:
                # Previously, this function would break hard here, because a
                # job cannot be immortalized twice. The thing is, however, that
                # sometimes a user moves the resource files to a different hard
                # drive or folder and then the upload job cannot be summoned
                # (The immortalized job has the wrong resource paths)! We offer
                # the user a workaround: If the dataset_id is in the eternal
                # job list, but not in self (because summoning it failed),
                # then we remove the old task from the eternal job list and
                # add the new task.
                if self.jobs_eternal.is_job_queued(upload_job.dataset_id):
                    try:
                        self.get_job(upload_job.dataset_id)
                    except KeyError:
                        # Job is immortalized, but failed to be summoned to
                        # Queue during initialization.
                        self.jobs_eternal.obliterate_job(upload_job.dataset_id)
                    else:
                        # Job is immortalized and already in the queue.
                        # Everything is fine and we need not worry. Must not
                        # append job!
                        return "known"
                else:
                    # Job is immortalized and already done! Under no
                    # circumstance should we add this job.
                    return "finished"
            # Add to eternal jobs for persistence
            self.jobs_eternal.immortalize_job(upload_job)
        self.jobs.append(upload_job)
        return "new"

    def get_job(self, dataset_id):
        """Return the queued UploadJob belonging to the dataset ID"""
        for job in self.jobs:
            if job.dataset_id == dataset_id:
                return job
        else:
            raise KeyError("Job '{}' not found!".format(dataset_id))

    def get_status(self, dataset_id):
        """Return the status of an UploadJob"""
        self.get_job(dataset_id).get_status()

    def new_job(self, dataset_id, paths, resource_names=None,
                supplements=None):
        """Create an UploadJob and add it to the upload queue

        Parameters
        ----------
        dataset_id: str
            The CKAN/DCOR dataset ID
        paths: list of str or list of pathlib.Path
            Paths to the resource to upload
        resource_names: list of str
            The names under which the resources are stored
        supplements: list of dict
            Resource schema supplements

        Returns
        -------
        upload_job: dcoraid.upload.job.UploadJob
            The upload job that was appended to the upload queue
        """
        upload_job = UploadJob(
            api=self.api,
            dataset_id=dataset_id,
            resource_paths=paths,
            resource_names=resource_names,
            resource_supplements=supplements,
            cache_dir=self.cache_dir,
        )
        self.add_job(upload_job)
        return upload_job

    def remove_job(self, dataset_id):
        """Remove an UploadJob from the queue and perform cleanup

        It has not been tested what happens when a running job
        is aborted. It will probably keep running and then complain
        about resources that are expected to be there. Don't do it.
        """
        for ii, job in enumerate(list(self.jobs)):
            if job.dataset_id == dataset_id:
                self.jobs.pop(ii)
                job.cleanup()
        # also remove from eternal jobs
        if (self.jobs_eternal is not None
                and self.jobs_eternal.is_job_queued(dataset_id)):
            self.jobs_eternal.obliterate_job(dataset_id)


class CompressDaemon(Daemon):
    def __init__(self, jobs):
        """Compression daemon"""
        super(CompressDaemon, self).__init__(
            jobs,
            job_trigger_state="init",
            job_function_name="task_compress_resources")


class UploadDaemon(Daemon):
    def __init__(self, jobs):
        """Upload daemon"""
        super(UploadDaemon, self).__init__(
            jobs,
            job_trigger_state="parcel",
            job_function_name="task_upload_resources")


class VerifyDaemon(Daemon):
    def __init__(self, jobs):
        """Verify daemon"""
        super(VerifyDaemon, self).__init__(
            jobs,
            job_trigger_state="online",
            job_function_name="task_verify_resources")



import abc

from ..common import ttl_cache


class DBInterrogator(abc.ABC):
    def __init__(self, mode, user_data):
        if mode == "user":
            mandatory = ["id", "name", "number_created_packages"]
            missing = [key for key in mandatory if key not in user_data]
            if missing:
                raise ValueError("The following keys are missing in "
                                 f"`user_data` for `mode=='user'`: {missing}")
        self.user_data = user_data
        self.mode = mode
        self.search_query = {}

    @property
    def is_up_to_date(self):
        """Checks whether the local database copy is up-to-date"""
        if self.local_version_score != self.remote_version_score:
            uptodate = False
        elif self.local_timestamp != self.remote_timestamp:
            uptodate = False
        else:
            uptodate = True
        return uptodate

    @abc.abstractmethod
    def get_circles(self):
        """Return the list of DCOR Circles"""
        pass

    @abc.abstractmethod
    def get_collections(self):
        """Return the list of DCOR Collections"""
        pass

    def get_datasets_user(self):
        """Return DBExtract with data owned by or shared with the user"""
        if self.mode != "user":
            raise ValueError("Cannot get user datasets if mode is 'public'!")
        owned = self.get_datasets_user_owned()
        shared = self.get_datasets_user_shared()
        following = self.get_datasets_user_following()
        # these are all instances of DBExtract
        return owned + shared + following

    @abc.abstractmethod
    def get_datasets_user_following(self):
        """Return all datasets the user is following"""
        pass

    @abc.abstractmethod
    def get_datasets_user_owned(self):
        """Return all datasets owned by the user"""
        pass

    @abc.abstractmethod
    def get_datasets_user_shared(self):
        """Return all datasets shared with the user"""
        pass

    @abc.abstractmethod
    def get_users(self):
        """Return the list of DCOR users"""
        pass

    @abc.abstractmethod
    def search_dataset(self, query, circles=None, collections=None):
        pass

    @property
    @abc.abstractmethod
    def local_version_score(self):
        """Local database version"""

    @property
    @abc.abstractmethod
    def local_timestamp(self):
        """Local database date in seconds since epoch"""

    @property
    @ttl_cache(seconds=5)
    def remote_version_score(self):
        """Remote database version"""
        return 0

    @property
    @ttl_cache(seconds=5)
    def remote_timestamp(self):
        """Remote database date in seconds since epoch"""
        return 0



import copy
import functools
import pathlib
import shutil
import time
import warnings

import requests

from ..common import sha256sum


#: Valid job states (in more or less chronological order)
JOB_STATES = [
    "init",  # initial
    "wait-disk",  # waiting for free disk space to download
    "transfer",  # download in progress
    "downloaded",  # dataset has been transferred
    "verify",  # perform SHA256 sum verification
    "done",  # job finished
    "abort",  # user aborted
    "error",  # error occurred
]


class DownloadJob(object):
    def __init__(self, api, resource_id, download_path):
        """Wrapper for resource downloads

        The function `task_download_resource` performs the actual
        download. During download, the progress is monitored and
        can be read from other threads using the e.g. `get_status`
        function.

        Parameters
        ----------
        api: dcoraid.api.CKANAPI
            The CKAN/DCOR API instance used for the download
        resource_id: str
            ID of the CKAN/DCOR resource
        download_path: str or pathlib.Path
            Local path to where the resource will be downloaded.
            If a directory is specified, that directory must exist.
            If the path to the local target is specified, its
            parent directory must exist.
        """
        self.api = api.copy()  # create a copy of the API
        self.resource_id = resource_id
        self.path = pathlib.Path(download_path)
        self.path_temp = None
        self.state = None
        self.set_state("init")
        self.traceback = None
        #: The bytes of the file that has been downloaded in the current
        #: session. It does not include bytes from a previous session
        #: (after resuming a download).
        self.file_bytes_downloaded = 0
        self.start_time = None
        self.end_time = None
        self._last_time = 0
        self._last_bytes = 0
        self._last_rate = 0

    def __getstate__(self):
        """Get the state of the DownloadJob instance

        This is not the state of the download! For monitoring the
        download, please see :func:`DownloadJob.get_status`,
        :const:`DownloadJob.state`, and :func:`DownloadJob.set_state`.

        See Also
        --------
        from_download_job_state: to recreate a job from a state
        """
        dj_state = {
            "resource_id": self.resource_id,
            "download_path": str(self.path),
        }
        return dj_state

    @property
    def file_size(self):
        return self.get_resource_dict()["size"]

    @staticmethod
    def from_download_job_state(dj_state, api):
        """Reinstantiate a job from an `DownloadJob.__getstate__` dict
        """
        return DownloadJob(api=api, **dj_state)

    @functools.lru_cache(maxsize=100)
    def get_resource_dict(self):
        """Return resource dictionary"""
        return self.api.get("resource_show", id=self.resource_id)

    @functools.lru_cache(maxsize=100)
    def get_dataset_dict(self):
        res_dict = self.get_resource_dict()
        ds_dict = self.api.get("package_show", id=res_dict["package_id"])
        return ds_dict

    def get_resource_url(self):
        """Return a link to the resource on DCOR"""
        ddict = self.get_resource_dict()
        return f"{self.api.server}/dataset/{ddict['package_id']}" \
               + f"/resource/{self.resource_id}/download/{ddict['name']}"

    def get_progress_string(self):
        """Return a nice string representation of the progress"""
        status = self.get_status()
        state = status["state"]

        if state in ["init", "transfer", "wait-disk"]:
            progress = "{:.0f}%".format(
                status["bytes local"]/status["bytes total"]*100)
        elif state in ["downloaded", "verify", "done"]:
            progress = "100%"
        elif state in ["abort", "error"]:
            progress = "--"
        elif state in JOB_STATES:
            # seems like you missed to update a case here?
            warnings.warn(f"Please add state '{state}' to these cases!")
            progress = "undefined"
        return progress

    def get_rate_string(self):
        """Return a nice string representing download rate"""
        status = self.get_status()
        state = status["state"]
        rate = status["rate"]
        if state in ["init", "wait-disk"]:
            rate_label = "-- kB/s"
        else:
            if rate > 1e6:
                rate_label = "{:.1f} MB/s".format(rate/1e6)
            else:
                rate_label = "{:.0f} kB/s".format(rate/1e3)
            if state != "transfer":
                rate_label = "⌀ " + rate_label
        return rate_label

    def get_rate(self, resolution=3.05):
        """Return the current resource download rate

        Parameters
        ----------
        resolution: float
            Time interval in which to perform the average.

        Returns
        -------
        download_rate: float
            Mean download rate in bytes per second
        """
        cur_time = time.perf_counter()
        # get bytes of files that have been downloaded
        cur_bytes = self.file_bytes_downloaded
        delta_time = cur_time - self._last_time
        delta_bytes = cur_bytes - self._last_bytes

        if self.start_time is None:
            # not started yet
            rate = 0
            self._last_bytes = 0
        elif self.end_time is None:
            # not finished yet
            if self._last_time == 0:
                # first time we are here
                delta_time = cur_time - self.start_time
            if delta_time > resolution:
                rate = delta_bytes / delta_time
                self._last_time = cur_time
                self._last_bytes = cur_bytes
            else:
                rate = self._last_rate
        else:
            # finished
            tdelt = (self.end_time - self.start_time)
            rate = cur_bytes / tdelt
        self._last_rate = rate
        return rate

    def get_status(self):
        """Get the status of the current job

        Returns
        -------
        status: dict
            Dictionary with various interesting parameters
        """
        data = {
            "state": self.state,
            "bytes total": self.file_size,
            "bytes downloaded": self.file_bytes_downloaded,
            "rate": self.get_rate(),
        }
        if self.path is not None and self.path.is_file():
            data["bytes local"] = self.path.stat().st_size
        elif self.path_temp is not None and self.path_temp.is_file():
            data["bytes local"] = self.path_temp.stat().st_size
        else:
            data["bytes local"] = 0

        return data

    def retry_download(self):
        """Retry downloading resources when an error occured"""
        if self.state in ["abort", "error"]:
            self.set_state("init")
        else:
            raise ValueError("Can only retry download in error state!")

    def set_state(self, state):
        """Set the current job state

        The state is checked against :const:`JOB_STATES`
        """
        if state not in JOB_STATES:
            raise ValueError("Unknown state: '{}'".format(state))
        self.state = state

    def task_download_resource(self):
        """Start the download

        The progress of the download is monitored and written
        to attributes. The current status can be retrieved
        via :func:`DownloadJob.get_status`.
        """
        if self.state in ["init", "wait-disk"]:
            # set-up temporary path
            if self.path.is_dir():
                # if a directory is given, we prepend the dataset name
                ds_name = self.get_dataset_dict()["name"]
                res_name = self.get_resource_dict()["name"]
                ds_dir = self.path / ds_name
                ds_dir.mkdir(exist_ok=True, parents=True)
                self.path = ds_dir / res_name
            if self.path.exists():
                self.start_time = None
                self.end_time = None
                self.file_bytes_downloaded = 0
                self.set_state("downloaded")
            else:
                self.path_temp = self.path.with_name(self.path.name + "~")
                # check for disk space
                size = self.get_resource_dict()["size"]
                if shutil.disk_usage(self.path_temp.parent).free < size:
                    # there is not enough space on disk for the download
                    self.set_state("wait-disk")
                    time.sleep(.2)
                else:
                    # proceed with download
                    # reset everything
                    self.file_bytes_downloaded = 0
                    self.start_time = None
                    self.end_time = None
                    self._last_time = 0
                    self._last_bytes = 0
                    self._last_rate = 0
                    # begin transfer
                    self.set_state("transfer")
                    self.start_time = time.perf_counter()
                    # Do the things to do and watch self.state while doing so
                    url = self.get_resource_url()
                    headers = copy.deepcopy(self.api.headers)
                    if self.path_temp.exists():
                        # resume previous download
                        bytes_present = self.path_temp.stat().st_size
                        headers["Range"] = f"bytes={bytes_present}-"
                    with requests.get(url, stream=True, headers=headers) as r:
                        r.raise_for_status()
                        with self.path_temp.open('ab') as f:
                            chunk_size = 1024*1024
                            for chunk in r.iter_content(chunk_size=chunk_size):
                                # If you have chunk encoded response uncomment
                                # if and set chunk_size parameter to None.
                                # if chunk:
                                f.write(chunk)
                                self.file_bytes_downloaded += len(chunk)
                    self.end_time = time.perf_counter()
                    self.set_state("downloaded")
        else:
            warnings.warn("Starting a download only possible when state is "
                          + "'init' or 'wait-disk', but current state is "
                          + "'{}'!".format(self.state))

    def task_verify_resource(self):
        """Perform SHA256 verification"""
        if self.state == "downloaded":
            if self.path is not None and self.path.is_file():
                if self.path_temp is not None and self.path_temp.exists():
                    self.path_temp.unlink()
                self.set_state("done")
            else:  # only verify if we have self.temp_path
                self.set_state("verify")
                # First check whether all SHA256 sums are already available
                # online
                if (self.get_resource_dict()["sha256"]
                        != sha256sum(self.path_temp)):
                    self.set_state("error")
                    self.traceback = "SHA256 sum check failed!"
                else:
                    self.path_temp.rename(self.path)
                    self.set_state("done")
        elif self.state != "done":  # ignore state "done" [sic!]
            # Only issue this warning if the download is not already done.
            warnings.warn("Resource verification is only possible when state "
                          + "is 'downloaded', but current state is "
                          + "'{}'!".format(self.state))



import traceback

import time

from ..common import ConnectionTimeoutErrors

from .kthread import KThread


class Daemon(KThread):
    def __init__(self, queue, job_trigger_state, job_function_name):
        """Daemon base class for running uploads/downloads in the background"""
        self.queue = queue
        self.state = "running"
        self.job_trigger_state = job_trigger_state
        self.job_function_name = job_function_name
        super(Daemon, self).__init__()
        self.daemon = True  # We don't have to worry about ending this thread
        self.start()

    def join(self, *args, **kwargs):
        """Join thread by breaking the while loop"""
        self.state = "exiting"
        super(Daemon, self).join(*args, **kwargs)
        assert self.state == "exited"

    def run(self):
        while True:
            if self.state == "exiting":
                self.state = "exited"
                break
            elif self.state != "running":
                # Don't do anything
                time.sleep(.1)
                continue
            else:
                # Get the first job that is in the trigger state
                for job in self.queue:
                    if job.state == self.job_trigger_state:
                        break
                else:
                    # Nothing to do, sleep a little to avoid 100% CPU
                    time.sleep(.1)
                    continue
                # Perform daemon task
                task = getattr(job, self.job_function_name)
                try:
                    task()
                except ConnectionTimeoutErrors:
                    # Set the job to the error state for 10s (so the user
                    # sees it in the UI) and then go back to the initial
                    # job trigger state.
                    job.set_state("error")
                    job.traceback = traceback.format_exc(limit=1) \
                        + "\nDCOR-Aid will retry in 10s!"
                    time.sleep(10)
                    job.set_state(self.job_trigger_state)
                except SystemExit:
                    job.set_state("abort")
                except BaseException:
                    # Set job to error state and let the user figure
                    # out what to do next.
                    job.set_state("error")
                    job.traceback = traceback.format_exc()



import tempfile
import pathlib
import shutil
import time
import warnings

from dclab.rtdc_dataset.check import IntegrityChecker
from dclab.cli import compress

from ..api import (dataset_activate, resource_add, resource_exists,
                   resource_sha256_sums)
from ..common import sha256sum


#: Valid job states (in more or less chronological order)
JOB_STATES = [
    "init",  # initial
    "wait-disk",  # waiting for free disk space to compress
    "compress",  # compression (only for DC data)
    "parcel",  # ready for upload
    "transfer",  # upload in progress
    "online",  # dataset has been transferred
    "verify",  # perform SHA256 sum verification
    "finalize",  # CKAN dataset is being activated
    "done",  # job finished
    "abort",  # user aborted
    "error",  # error occurred
]


class UploadJob(object):
    def __init__(self, api, dataset_id, resource_paths,
                 resource_names=None, resource_supplements=None,
                 task_id=None, cache_dir=None):
        """Wrapper for resource uploads

        This job is meant to be run from a separate thread.
        The function `task_compress_resources` ensures that all
        resources are compressed. The function `task_upload_resources`
        performs the actual upload. During upload, the progress
        is monitored and can be read from other threads
        using the e.g. `get_status` function.

        Parameters
        ----------
        api: dcoraid.api.CKANAPI
            The CKAN/DCOR API instance used for the upload
        dataset_id: str
            ID of the CKAN/DCOR dataset
        resource_paths: list
            Paths to the dataset resources
        resource_names: list
            Corresponding names of the resources as they should appear
            on DCOR
        resource_supplements: list of dict
            Supplementary resource information
        task_id: str
            Unique task ID (used for identifying jobs uploaded already)
        cache_dir: str or pathlib.Path
            Cache directory for storing compressed .rtdc files;
            if not supplied, a temporary directory is created
        """
        self.api = api.copy()  # create a copy of the API
        self.dataset_id = dataset_id
        # make sure the dataset_id is valid
        self.api.get("package_show", id=self.dataset_id)
        self.paths = [pathlib.Path(pp).resolve() for pp in resource_paths]
        if resource_names is None:
            resource_names = [pp.name for pp in self.paths]
        #: Important supplementary resource schema meta data that will
        #: be formatted to composite {"sp:section:key" = value} an appended
        #: to the resource metadata.
        self.supplements = resource_supplements
        self.resource_names = resource_names
        self.task_id = task_id
        self.paths_uploaded = []
        self.paths_uploaded_before = []
        self.state = None
        self.set_state("init")
        self.traceback = None
        self.file_sizes = [pathlib.Path(ff).stat().st_size
                           for ff in self.paths]
        self.file_bytes_uploaded = [0] * len(self.paths)
        self.index = 0
        self.start_time = None
        self.end_time = None
        self._last_time = 0
        self._last_bytes = 0
        self._last_rate = 0
        # caching
        if cache_dir is None:
            cache_dir = pathlib.Path(tempfile.mkdtemp(
                prefix=f"dcoraid_upload_compress-{self.dataset_id}_"))
        else:
            cache_dir = pathlib.Path(cache_dir) / f"compress-{self.dataset_id}"
        self.cache_dir = cache_dir

    def __getstate__(self):
        """Get the state of the UploadJob instance

        This is not the state of the upload! For monitoring the
        upload, please see :func:`UploadJob.get_status`,
        :const:`UploadJob.state`, and :func:`UploadJob.set_state`.

        See Also
        --------
        from_upload_job_state: to recreate a job from a state
        """
        uj_state = {
            "dataset_id": self.dataset_id,
            "resource_paths": [str(pp) for pp in self.paths],
            "resource_names": self.resource_names,
            "resource_supplements": self.supplements,
            "task_id": self.task_id,
        }
        return uj_state

    @staticmethod
    def from_upload_job_state(uj_state, api, cache_dir=None):
        """Reinstantiate a job from an `UploadJob.__getstate__` dict

        Note that there is no `UploadJob.__setstate__` function,
        because that would just not be possible/debugabble when one
        of the tasks (compress, upload) is running. The upload job
        is implemented in such a way that it will just skip existing
        resources, so it will just go through the states until it
        is "done", even if there is nothing to upload. BTW this is
        also a great way of making sure that an upload is complete.

        Note
        ----
        The `uj_state` dictionary must contain the "dataset_id",
        otherwise we won't know where to upload the resources to.
        """
        return UploadJob(api=api, cache_dir=cache_dir, **uj_state)

    def cleanup(self):
        """cleanup temporary files in the user's cache directory"""
        shutil.rmtree(self.cache_dir, ignore_errors=True)

    def get_composite_supplements(self, resource_index):
        """Return the composite supplements "sp:section:key" for a resource"""
        sp = {}
        if self.supplements:
            sp_dict = self.supplements[resource_index]
            for sec in sp_dict:
                for key in sp_dict[sec]:
                    sp["sp:{}:{}".format(sec, key)] = sp_dict[sec][key]
        return sp

    def get_dataset_url(self):
        """Return a link to the dataset on DCOR"""
        # The API prepends missing "https://"
        return "{}/dataset/{}".format(self.api.server, self.dataset_id)

    def get_progress_string(self):
        """Return a nice string representation of the progress"""
        status = self.get_status()
        state = status["state"]
        plural = "s" if status["files total"] > 1 else ""

        if state in ["init", "compress", "parcel", "wait-disk"]:
            progress = "0% ({} file{})".format(status["files total"], plural)
        elif state == "transfer":
            progress = "{:.0f}% (file {}/{})".format(
                status["bytes uploaded"]/status["bytes total"]*100,
                status["files uploaded"]+1,
                status["files total"])
        elif state in ["finalize", "online", "verify", "done"]:
            progress = "100% ({} file{})".format(status["files total"],
                                                 plural)
        elif state in ["abort", "error"]:
            progress = "-- ({}/{} file{})".format(status["files uploaded"],
                                                  status["files total"],
                                                  plural)
        elif state in JOB_STATES:
            # seems like you missed to update a case here?
            warnings.warn(f"Please add state '{state}' to these cases!")
            progress = "undefined"
        return progress

    def get_rate_string(self):
        """Return a nice string representing upload rate"""
        status = self.get_status()
        state = status["state"]
        rate = status["rate"]
        if state in ["init", "compress", "parcel"]:
            rate_label = "-- kB/s"
        else:
            if rate > 1e6:
                rate_label = "{:.1f} MB/s".format(rate/1e6)
            else:
                rate_label = "{:.0f} kB/s".format(rate/1e3)
            if state != "transfer":
                rate_label = "⌀ " + rate_label
        return rate_label

    def get_rate(self, resolution=3.05):
        """Return the current resource upload rate

        Parameters
        ----------
        resolution: float
            Time interval in which to perform the average.

        Returns
        -------
        upload_rate: float
            Mean upload rate in bytes per second
        """
        cur_time = time.perf_counter()
        # get bytes of files that have been uploaded
        cur_bytes = 0
        for ii, path in enumerate(self.paths):
            if path not in self.paths_uploaded_before:
                # only count files from the current session
                cur_bytes += self.file_bytes_uploaded[ii]

        delta_time = cur_time - self._last_time
        delta_bytes = cur_bytes - self._last_bytes

        if self.start_time is None:
            # not started yet
            rate = 0
            self._last_bytes = 0
        elif self.end_time is None:
            # not finished yet
            if self._last_time == 0:
                # first time we are here
                delta_time = cur_time - self.start_time
            if delta_time > resolution:
                rate = delta_bytes / delta_time
                self._last_time = cur_time
                self._last_bytes = cur_bytes
            else:
                rate = self._last_rate
        else:
            # finished
            tdelt = (self.end_time - self.start_time)
            rate = cur_bytes / tdelt
        self._last_rate = rate
        return rate

    def get_status(self):
        """Get the status of the current job

        Returns
        -------
        status: dict
            Dictionary with various interesting parameters
        """
        data = {
            "state": self.state,
            "files total": len(self.paths),
            "files uploaded": len(self.paths_uploaded),
            "bytes total": sum(self.file_sizes),
            "bytes uploaded": sum(self.file_bytes_uploaded),
            "rate": self.get_rate(),
        }
        return data

    def monitor_callback(self, monitor):
        """Upload progress monitor callback function

        This method updates `self.file_sizes` and
        `self.file_bytes_uploaded` for the currently
        uploading resource.
        """
        self.file_sizes[self.index] = monitor.len
        self.file_bytes_uploaded[self.index] = monitor.bytes_read

    def retry_upload(self):
        """Retry uploading resources when an error occured"""
        if self.state in ["abort", "error"]:
            self.set_state("init")
        else:
            raise ValueError("Can only retry upload in error state!")

    def set_state(self, state):
        """Set the current job state

        The state is checked against :const:`JOB_STATES`
        """
        if state not in JOB_STATES:
            raise ValueError("Unknown state: '{}'".format(state))
        self.state = state

    def task_compress_resources(self):
        """Compress resources if they are not fully compressed

        Data are stored in the user's cache directory and
        deleted after upload is complete.
        """
        self.set_state("compress")
        for ii, path in enumerate(list(self.paths)):
            if path.suffix in [".rtdc", ".dc"]:  # do we have an .rtdc file?
                with IntegrityChecker(path) as ic:
                    # check for compression
                    cdata = ic.check_compression()[0].data
                    # perform a sanity check
                    insane = [c.msg for c in ic.sanity_check()]
                    if insane:
                        # The user is responsible of cleaning up the mess.
                        # We just make sure no dirty data gets uploaded to
                        # DCOR.
                        raise IOError(f"Sanity check failed for {path}!\n"
                                      + "\n".join(insane))
                if cdata["uncompressed"]:  # (partially) not compressed?
                    res_dir = self.cache_dir / "{}".format(ii)
                    res_dir.mkdir(exist_ok=True, parents=True)
                    path_out = res_dir / path.name
                    # Check if the output path already exists:
                    # We can trust dclab version 0.34.4 that the
                    # compression was successful, because it uses
                    # temporary file paths and only renames the
                    # temporary files to the output files when
                    # the conversion was successful.
                    if not path_out.exists():
                        # perform compression
                        size_in = path.stat().st_size
                        while shutil.disk_usage(res_dir).free < size_in:
                            # OK, now there might be the possibility that
                            # there are rogue upload caches from jobs that
                            # were created during previous runs. Since
                            # DCOR-Aid 0.5.4, upload jobs are run in
                            # alphabetical order. This means that we may
                            # delete the upload cache of jobs that would
                            # be uploaded after this job here.
                            # This is important so that for slow uploads,
                            # the hard disk is not filled up and then no more
                            # jobs can be uploaded because the existing cache
                            # data blocks jobs that are earlier in the queue
                            # from compressing themselves.
                            # find all upload job cache candidates
                            cands = sorted(
                                self.cache_dir.parent.glob("compress-*"))
                            # determine the index of the current curve
                            for ci, cand in enumerate(cands):
                                if cand.samefile(self.cache_dir):
                                    # we have found ourself
                                    for rogue in cands[ci+1:]:
                                        # Remove rogue cache data
                                        shutil.rmtree(rogue,
                                                      ignore_errors=True)
                                    # don't continue
                                    break

                            # As long as there is less space free than the
                            # input file size, we stall here.
                            self.set_state("wait-disk")
                            time.sleep(0.2)
                        self.set_state("compress")
                        compress(path_out=path_out, path_in=path)
                    # replace current path_out with compressed path
                    self.paths[ii] = path_out
        self.set_state("parcel")

    def task_upload_resources(self):
        """Start the upload

        The progress of the upload is monitored and written
        to attributes. The current status can be retrieved
        via :func:`UploadJob.get_status`.
        """
        if self.state == "parcel":
            # reset everything
            self.paths_uploaded.clear()
            self.paths_uploaded_before.clear()
            self.file_bytes_uploaded = [0] * len(self.paths)
            self.index = 0
            self.start_time = None
            self.end_time = None
            self._last_time = 0
            self._last_bytes = 0
            self._last_rate = 0
            # begin transfer
            self.set_state("transfer")
            self.start_time = time.perf_counter()
            # Do the things to do and watch self.state while doing so
            for ii, path in enumerate(self.paths):
                self.index = ii
                resource_name = self.resource_names[ii]
                resource_supplement = self.get_composite_supplements(ii)
                exists = resource_exists(
                    dataset_id=self.dataset_id,
                    resource_name=resource_name,
                    resource_dict=resource_supplement,
                    api=self.api)
                if exists:
                    # We are currently retrying an upload. If the
                    # resource exists, we have already uploaded it.
                    self.file_bytes_uploaded[ii] = self.file_sizes[ii]
                    self.paths_uploaded.append(path)
                    self.paths_uploaded_before.append(path)
                    continue
                else:
                    # Normal upload.
                    resource_add(
                        dataset_id=self.dataset_id,
                        path=path,
                        resource_name=resource_name,
                        resource_dict=resource_supplement,
                        api=self.api,
                        exist_ok=True,
                        monitor_callback=self.monitor_callback)
                    self.paths_uploaded.append(path)
                    # Workaround until `package_revise` is in `resource_add`
                    # (https://github.com/DCOR-dev/DCOR-Aid/issues/28):
                    # Wait for the SHA256 sum to be computed by DCOR.
                    for _ in range(60):
                        sha256dict = resource_sha256_sums(
                            dataset_id=self.dataset_id,
                            api=self.api)
                        if sha256dict[resource_name] is not None:
                            break
                        time.sleep(10)
                    else:
                        warnings.warn(
                            f"SHA256 sum for {self.dataset_id}: "
                            + f"{resource_name} not yet computed by DCOR!")
            self.end_time = time.perf_counter()
            self.set_state("online")
        else:
            warnings.warn("Starting an upload only possible when state is "
                          + "'parcel', but current state is "
                          + "'{}'!".format(self.state))

    def task_verify_resources(self):
        """Perform SHA256 verification"""
        if self.state == "online":
            # First check whether all SHA256 sums are already available online
            sha256dict = resource_sha256_sums(
                dataset_id=self.dataset_id,
                api=self.api)
            if sum([sha256dict[name] is None for name in sha256dict]) != 0:
                # only start verification if all SHA256 sums are available
                pass
            else:
                self.set_state("verify")
                bad_sha256 = []
                for ii, path in enumerate(self.paths):
                    resource_name = self.resource_names[ii]
                    # compute SHA256 sum
                    sha = sha256sum(path)
                    if sha != sha256dict[resource_name]:
                        bad_sha256.append(
                            [resource_name, sha, sha256dict[resource_name]])
                if bad_sha256:
                    # we have bad resources, tell the user
                    self.set_state("error")
                    msg_parts = ["SHA256 sum failed for resources:"]
                    for item in bad_sha256:
                        msg_parts.append("'{}' ({} vs. {})!".format(*item))
                    self.traceback = "\n".join(msg_parts)
                else:
                    # green means go: remove temporary files and activate
                    self.cleanup()
                    # finalize dataset
                    self.set_state("finalize")
                    # draft -> active
                    dataset_activate(
                        dataset_id=self.dataset_id,
                        api=self.api)
                    self.set_state("done")
        elif self.state != "done":  # ignore state "done" [sic!]
            # Only issue this warning if the upload is not already done.
            warnings.warn("Resource verification is only possible when state "
                          + "is 'online', but current state is "
                          + "'{}'!".format(self.state))



import pathlib
import tempfile
from dcoraid.download import job, task

import common


dpath = pathlib.Path(__file__).parent / "data" / "calibration_beads_47.rtdc"


def test_save_load_basic():
    api = common.get_api()
    td = tempfile.mkdtemp(prefix="test-download")
    task_path = pathlib.Path(td) / "test.json"
    ds_dict = common.make_dataset_for_download()
    dj = job.DownloadJob(api=api,
                         resource_id=ds_dict["resources"][0]["id"],
                         download_path=td)
    task.save_task(dj, task_path)

    dj2 = task.load_task(task_path, api=api)
    assert dj2.resource_id == dj.resource_id
